import asyncio
import os
import sys
import tempfile
import typing as t
from contextvars import ContextVar
from datetime import datetime
from enum import Enum
from functools import lru_cache
from importlib import import_module, util
from inspect import stack
from pathlib import Path
from uuid import UUID

# Removed nest_asyncio import - not needed in library code
import rich.repr
from anyio import Path as AsyncPath
from inflection import camelize
from msgspec.yaml import decode as yaml_decode
from pydantic import BaseModel, ConfigDict, Field
from acb.actions.encode import yaml_encode
from acb.depends import depends

# Declare uuid_utils to handle both import cases
uuid_utils: t.Any = None  # Initialize as None

try:
    import uuid_utils  # type: ignore[no-redef]

    _uuid7_available = True
    uuid_lib: t.Any = uuid_utils  # type: ignore[no-redef]
except ImportError:
    import uuid

    uuid_utils = None  # type: ignore[no-redef]

    _uuid7_available = False
    uuid_lib: t.Any = uuid  # type: ignore[no-redef]


class AdapterStatus(str, Enum):
    ALPHA = "alpha"
    BETA = "beta"
    STABLE = "stable"
    DEPRECATED = "deprecated"
    EXPERIMENTAL = "experimental"


class AdapterCapability(str, Enum):
    CONNECTION_POOLING = "connection_pooling"
    RECONNECTION = "auto_reconnection"
    HEALTH_CHECKS = "health_checks"
    TLS_SUPPORT = "tls_support"

    TRANSACTIONS = "transactions"
    BULK_OPERATIONS = "bulk_operations"
    STREAMING = "streaming"
    COMPRESSION = "compression"
    ENCRYPTION = "encryption"

    CACHING = "caching"
    ASYNC_OPERATIONS = "async_operations"
    BATCHING = "batching"

    METRICS = "metrics"
    TRACING = "tracing"
    LOGGING = "structured_logging"

    SCHEMA_VALIDATION = "schema_validation"
    MIGRATIONS = "migrations"
    BACKUP_RESTORE = "backup_restore"

    # Logger-specific capabilities
    STRUCTURED_OUTPUT = "structured_output"
    ASYNC_LOGGING = "async_logging"
    CONTEXTUAL = "contextual"
    ROTATION = "rotation"
    REMOTE_LOGGING = "remote_logging"
    JSON_OUTPUT = "json_output"
    FILE_LOGGING = "file_logging"
    CORRELATION_ID = "correlation_id"

    # Graph-specific capabilities
    GRAPH_TRAVERSAL = "graph_traversal"
    CYPHER_QUERIES = "cypher_queries"
    GREMLIN_QUERIES = "gremlin_queries"
    AQL_QUERIES = "aql_queries"
    GRAPH_ANALYTICS = "graph_analytics"
    PATHFINDING = "pathfinding"
    SUBGRAPH_OPERATIONS = "subgraph_operations"
    GRAPH_SCHEMA_VALIDATION = "graph_schema_validation"

    # AI/ML-specific capabilities
    HYBRID_DEPLOYMENT = "hybrid_deployment"
    EDGE_INFERENCE = "edge_inference"
    MULTIMODAL_PROCESSING = "multimodal_processing"
    PROMPT_TEMPLATING = "prompt_templating"
    MODEL_CACHING = "model_caching"
    ADAPTIVE_ROUTING = "adaptive_routing"
    FALLBACK_MECHANISMS = "fallback_mechanisms"
    TEXT_GENERATION = "text_generation"
    VISION_PROCESSING = "vision_processing"
    AUDIO_PROCESSING = "audio_processing"
    MODEL_QUANTIZATION = "model_quantization"
    COLD_START_OPTIMIZATION = "cold_start_optimization"

    # Embedding-specific capabilities
    BATCH_EMBEDDING = "batch_embedding"
    EDGE_OPTIMIZED = "edge_optimized"
    TEXT_PREPROCESSING = "text_preprocessing"
    VECTOR_NORMALIZATION = "vector_normalization"
    DIMENSION_SCALING = "dimension_scaling"
    SEMANTIC_SEARCH = "semantic_search"
    SIMILARITY_COMPUTATION = "similarity_computation"
    DOCUMENT_CHUNKING = "document_chunking"
    POOLING_STRATEGIES = "pooling_strategies"
    MEMORY_EFFICIENT_PROCESSING = "memory_efficient_processing"


class AdapterMetadata(BaseModel):
    module_id: UUID = Field(
        description="UUID7 identifier for this specific adapter module",
    )

    name: str = Field(description="Human-readable adapter name")
    category: str = Field(description="Adapter category (cache, sql, storage, etc.)")
    provider: str = Field(description="Technology provider (redis, mysql, s3, etc.)")

    version: str = Field(
        description="Semantic version of this adapter (independent of ACB version)",
    )
    acb_min_version: str = Field(description="Minimum ACB version required")
    acb_max_version: str | None = Field(
        default=None,
        description="Maximum ACB version supported (None = no limit)",
    )

    author: str = Field(description="Primary author/maintainer")
    created_date: str = Field(description="ISO date when adapter was created")
    last_modified: str = Field(description="ISO date of last significant update")

    status: AdapterStatus = Field(description="Development/stability status")
    capabilities: list[AdapterCapability] = Field(
        default_factory=list,
        description="List of features this adapter supports",
    )

    required_packages: list[str] = Field(
        default_factory=list,
        description="External packages required for this adapter",
    )
    optional_packages: dict[str, str] = Field(
        default_factory=dict,
        description="Optional packages and their purpose",
    )

    description: str = Field(description="Brief description of adapter functionality")
    documentation_url: str | None = Field(
        default=None,
        description="Link to detailed documentation",
    )
    repository_url: str | None = Field(
        default=None,
        description="Source code repository for this adapter",
    )

    settings_class: str = Field(description="Name of the settings class")
    config_example: dict[str, t.Any] | None = Field(
        default=None,
        description="Example configuration for this adapter",
    )

    custom: dict[str, t.Any] = Field(
        default_factory=dict,
        description="Custom metadata fields for specific requirements",
    )

    model_config = ConfigDict(use_enum_values=True, extra="forbid")


def generate_adapter_id() -> UUID:
    if _uuid7_available:
        uuid_obj = uuid_lib.uuid7()  # type: ignore[attr-defined]
        return (
            UUID(str(uuid_obj))
            if not isinstance(uuid_obj, UUID)
            else UUID(str(uuid_obj))
        )
    uuid_obj = uuid_lib.uuid4()
    # Explicitly cast to UUID to satisfy type checker
    return UUID(str(uuid_obj))


def create_metadata_template(
    name: str,
    category: str,
    provider: str,
    author: str,
    description: str,
    **kwargs: t.Any,
) -> AdapterMetadata:
    now = datetime.now().isoformat()

    return AdapterMetadata(
        module_id=generate_adapter_id(),
        name=name,
        category=category,
        provider=provider,
        version="1.0.0",
        acb_min_version="0.18.0",
        author=author,
        created_date=now,
        last_modified=now,
        status=AdapterStatus.ALPHA,
        description=description,
        settings_class=f"{category.title()}Settings",
        **kwargs,
    )


def validate_version_compatibility(
    adapter_metadata: AdapterMetadata,
    current_acb_version: str,
) -> bool:
    try:
        from packaging import version

        current = version.parse(current_acb_version)
        min_version = version.parse(adapter_metadata.acb_min_version)

        if current < min_version:
            return False

        if adapter_metadata.acb_max_version:
            max_version = version.parse(adapter_metadata.acb_max_version)
            if current > max_version:
                return False

        return True
    except ImportError:
        return current_acb_version >= adapter_metadata.acb_min_version


def extract_metadata_from_module(module: t.Any) -> AdapterMetadata | None:
    if hasattr(module, "MODULE_METADATA"):
        metadata = module.MODULE_METADATA
        if isinstance(metadata, AdapterMetadata):
            return metadata
        if isinstance(metadata, dict):
            return AdapterMetadata(**metadata)
    return None


def extract_metadata_from_class(adapter_class: type) -> AdapterMetadata | None:
    if hasattr(adapter_class, "__module_metadata__"):
        metadata = adapter_class.__module_metadata__
        if isinstance(metadata, AdapterMetadata):
            return metadata
        if isinstance(metadata, dict):
            return AdapterMetadata(**metadata)
    return None


def get_adapter_info(adapter_class: type) -> dict[str, t.Any]:
    info: dict[str, t.Any] = {
        "class_name": adapter_class.__name__,
        "module_path": adapter_class.__module__,
        "has_metadata": False,
    }
    metadata = extract_metadata_from_class(adapter_class)
    if metadata:
        info["has_metadata"] = True
        info["module_id"] = str(metadata.module_id)
        info["name"] = metadata.name
        info["category"] = metadata.category
        info["provider"] = metadata.provider
        info["version"] = metadata.version
        info["status"] = (
            metadata.status.value
            if hasattr(metadata.status, "value")
            else metadata.status
        )
        info["author"] = metadata.author
        info["capabilities"] = [
            cap.value if hasattr(cap, "value") else cap for cap in metadata.capabilities
        ]
        info["required_packages"] = metadata.required_packages
        info["description"] = metadata.description
        info["acb_min_version"] = metadata.acb_min_version
        info["acb_max_version"] = metadata.acb_max_version

    return info


def check_adapter_capability(
    adapter_class: type,
    capability: AdapterCapability,
) -> bool:
    metadata = extract_metadata_from_class(adapter_class)
    if metadata:
        return capability in metadata.capabilities
    return False


def list_adapter_capabilities(adapter_class: type) -> list[str]:
    metadata = extract_metadata_from_class(adapter_class)
    if metadata:
        return [
            cap.value if hasattr(cap, "value") else cap for cap in metadata.capabilities
        ]
    return []


def generate_adapter_report(adapter_class: type) -> str:
    info = get_adapter_info(adapter_class)
    if not info["has_metadata"]:
        return f"Adapter: {info['class_name']} (No metadata available)"

    return f"""
Adapter Report: {info["name"]}
{"=" * (16 + len(info["name"]))}

Basic Information:
  Module ID: {info["module_id"]}
  Category: {info["category"]}
  Provider: {info["provider"]}
  Version: {info["version"]}
  Status: {info["status"]}
  Author: {info["author"]}

Compatibility:
  ACB Min Version: {info["acb_min_version"]}
  ACB Max Version: {info["acb_max_version"] or "No limit"}

Capabilities ({len(info["capabilities"])}):
  {chr(10).join("- " + cap for cap in info["capabilities"])}

Dependencies ({len(info["required_packages"])}):
  {chr(10).join("- " + pkg for pkg in info["required_packages"])}

Description:
  {info["description"]}

Module: {info["module_path"]}
Class: {info["class_name"]}
""".strip()


# Removed nest_asyncio.apply() - not needed in library code
_deployed: bool = os.getenv("DEPLOYED", "False").lower() == "true"
_testing: bool = os.getenv("TESTING", "False").lower() == "true"
root_path: AsyncPath = AsyncPath(Path.cwd())
tmp_path: AsyncPath = root_path / "tmp"
adapters_path: AsyncPath = root_path / "adapters"
actions_path: AsyncPath = root_path / "actions"
settings_path: AsyncPath = root_path / "settings"
app_settings_path: AsyncPath = settings_path / "app.yaml"
debug_settings_path: AsyncPath = settings_path / "debug.yaml"
adapter_settings_path: AsyncPath = settings_path / "adapters.yaml"
secrets_path: AsyncPath = (
    AsyncPath(Path(tempfile.mkdtemp(prefix="mock_secrets_")))
    if _testing or "pytest" in sys.modules
    else tmp_path / "secrets"
)
_install_lock: ContextVar[list[str]] = ContextVar("install_lock", default=[])
_adapter_import_locks: ContextVar[dict[str, asyncio.Lock]] = ContextVar(
    "_adapter_import_locks",
    default={},
)


class AdapterNotFound(Exception): ...


class AdapterNotInstalled(Exception): ...


class StaticImportError(ImportError): ...


AdapterClass = t.TypeVar("AdapterClass")


@t.runtime_checkable
class AdapterProtocol(t.Protocol):
    config: t.Any = None
    logger: t.Any = None

    async def init(self) -> None: ...


@rich.repr.auto
class Adapter(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str
    class_name: str
    category: str
    pkg: str = "acb"
    module: str = ""
    enabled: bool = False
    installed: bool = False
    path: AsyncPath = AsyncPath(__file__)

    metadata: AdapterMetadata | None = None
    runtime_id: str | None = None

    def __str__(self) -> str:
        return self.__repr__()

    def __hash__(self) -> int:
        base_hash = (self.name, self.class_name, self.category, self.pkg, self.module)
        if self.metadata:
            return hash((*base_hash, str(self.metadata.module_id)))
        return hash(base_hash)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Adapter):
            return False
        return (
            self.name == other.name
            and self.class_name == other.class_name
            and (self.category == other.category)
            and (self.pkg == other.pkg)
            and (self.module == other.module)
        )


adapter_registry: ContextVar[list[Adapter]] = ContextVar("adapter_registry", default=[])
_enabled_adapters_cache: ContextVar[dict[str, Adapter]] = ContextVar(
    "_enabled_adapters_cache",
    default={},
)
_installed_adapters_cache: ContextVar[dict[str, Adapter]] = ContextVar(
    "_installed_adapters_cache",
    default={},
)
core_adapters = [
    Adapter(
        name="config",
        module="acb.config",
        class_name="Config",
        category="config",
        enabled=True,
        installed=True,
        path=AsyncPath(__file__).parent / "config.py",
    ),
    Adapter(
        name="loguru",
        module="acb.adapters.logger.loguru",
        class_name="Logger",
        category="logger",
        enabled=True,
        installed=True,
        path=AsyncPath(__file__).parent / "logger" / "loguru.py",
    ),
]


def _update_adapter_caches() -> None:
    enabled_cache = {}
    installed_cache = {}
    for adapter in adapter_registry.get():
        if adapter.enabled:
            enabled_cache[adapter.category] = adapter
        if adapter.installed:
            installed_cache[adapter.category] = adapter
    _enabled_adapters_cache.get().update(enabled_cache)
    _installed_adapters_cache.get().update(installed_cache)


adapter_registry.get().extend([*core_adapters])
_update_adapter_caches()


def get_adapter(category: str) -> Adapter | None:
    return _enabled_adapters_cache.get().get(category)


def get_adapters() -> list[Adapter]:
    return list(_enabled_adapters_cache.get().values())


def get_installed_adapter(category: str) -> Adapter | None:
    return _installed_adapters_cache.get().get(category)


def get_installed_adapters() -> list[Adapter]:
    return list(_installed_adapters_cache.get().values())


# Static adapter mappings for improved performance
# This eliminates the need for dynamic string construction and reduces import overhead
STATIC_ADAPTER_MAPPINGS = {
    "cache.memory": ("acb.adapters.cache.memory", "Cache"),
    "cache.redis": ("acb.adapters.cache.redis", "Cache"),
    "logger": ("acb.adapters.logger.loguru", "Logger"),  # Default logger (loguru)
    "logger.loguru": ("acb.adapters.logger.loguru", "Logger"),
    "logger.structlog": ("acb.adapters.logger.structlog", "Logger"),
    "logger.logly": ("acb.adapters.logger.logly", "Logger"),
    "sql.mysql": ("acb.adapters.sql.mysql", "Sql"),
    "sql.pgsql": ("acb.adapters.sql.pgsql", "Sql"),
    "sql.sqlite": ("acb.adapters.sql.sqlite", "Sql"),
    "nosql.mongodb": ("acb.adapters.nosql.mongodb", "Nosql"),
    "nosql.firestore": ("acb.adapters.nosql.firestore", "Nosql"),
    "nosql.redis": ("acb.adapters.nosql.redis", "Nosql"),
    "storage.file": ("acb.adapters.storage.file", "Storage"),
    "storage.s3": ("acb.adapters.storage.s3", "Storage"),
    "storage.azure": ("acb.adapters.storage.azure", "Storage"),
    "storage.cloud_storage": ("acb.adapters.storage.cloud_storage", "Storage"),
    "storage.memory": ("acb.adapters.storage.memory", "Storage"),
    "secret.infisical": ("acb.adapters.secret.infisical", "Secret"),
    "secret.secret_manager": ("acb.adapters.secret.secret_manager", "Secret"),
    "secret.azure": ("acb.adapters.secret.azure", "Secret"),
    "secret.cloudflare": ("acb.adapters.secret.cloudflare", "Secret"),
    "monitoring.sentry": ("acb.adapters.monitoring.sentry", "Monitoring"),
    "monitoring.logfire": ("acb.adapters.monitoring.logfire", "Monitoring"),
    "requests.httpx": ("acb.adapters.requests.httpx", "Requests"),
    "requests.niquests": ("acb.adapters.requests.niquests", "Requests"),
    "smtp.gmail": ("acb.adapters.smtp.gmail", "Smtp"),
    "smtp.mailgun": ("acb.adapters.smtp.mailgun", "Smtp"),
    "dns.cloud_dns": ("acb.adapters.dns.cloud_dns", "Dns"),
    "dns.cloudflare": ("acb.adapters.dns.cloudflare", "Dns"),
    "dns.route53": ("acb.adapters.dns.route53", "Dns"),
    "ftpd.ftp": ("acb.adapters.ftpd.ftp", "Ftpd"),
    "ftpd.sftp": ("acb.adapters.ftpd.sftp", "Ftpd"),
    "models": ("acb.adapters.models", "Models"),
    "vector.duckdb": ("acb.adapters.vector.duckdb", "Vector"),
    "vector.pinecone": ("acb.adapters.vector.pinecone", "Vector"),
    "vector.qdrant": ("acb.adapters.vector.qdrant", "Vector"),
    "vector.weaviate": ("acb.adapters.vector.weaviate", "Vector"),
    "graph.neo4j": ("acb.adapters.graph.neo4j", "Graph"),
    "graph.neptune": ("acb.adapters.graph.neptune", "Graph"),
    "graph.arangodb": ("acb.adapters.graph.arangodb", "Graph"),
    # Pub/Sub adapters for events system
    "pubsub.memory": ("acb.adapters.messaging.memory", "MemoryPubSub"),
    "pubsub.redis": ("acb.adapters.messaging.redis", "RedisPubSub"),
    "pubsub.rabbitmq": ("acb.adapters.messaging.rabbitmq", "RabbitMQPubSub"),
    "pubsub.aiormq": ("acb.adapters.messaging.aiormq", "AioRmqPubSub"),
    # Queue adapters for tasks system
    "queue.memory": ("acb.adapters.messaging.memory", "MemoryQueue"),
    "queue.redis": ("acb.adapters.messaging.redis", "RedisQueue"),
    "queue.rabbitmq": ("acb.adapters.messaging.rabbitmq", "RabbitMQQueue"),
    "queue.aiormq": ("acb.adapters.messaging.aiormq", "AioRmqQueue"),
    # Task system adapters
    "tasks.memory": ("acb.tasks.memory", "TaskSystem"),
    "tasks.redis": ("acb.tasks.redis", "TaskSystem"),
    "tasks.rabbitmq": ("acb.tasks.rabbitmq", "TaskSystem"),
    "tasks.apscheduler": ("acb.tasks.apscheduler", "TaskSystem"),
    "ai.cloud": ("acb.adapters.ai.cloud", "CloudAI"),
    "ai.edge": ("acb.adapters.ai.edge", "EdgeAI"),
    "ai.hybrid": ("acb.adapters.ai.hybrid", "HybridAI"),
    "embedding.openai": ("acb.adapters.embedding.openai", "Embedding"),
    "embedding.huggingface": ("acb.adapters.embedding.huggingface", "Embedding"),
    "embedding.sentence_transformers": (
        "acb.adapters.embedding.sentence_transformers",
        "Embedding",
    ),
    "embedding.onnx": ("acb.adapters.embedding.onnx", "Embedding"),
    "embedding.lfm": ("acb.adapters.embedding.lfm", "Embedding"),
    "reasoning.langchain": ("acb.adapters.reasoning.langchain", "Reasoning"),
    "reasoning.llamaindex": ("acb.adapters.reasoning.llamaindex", "Reasoning"),
    "reasoning.custom": ("acb.adapters.reasoning.custom", "Reasoning"),
    "reasoning.openai_functions": (
        "acb.adapters.reasoning.openai_functions",
        "Reasoning",
    ),
}


@lru_cache(maxsize=256)
def get_adapter_class(category: str, name: str) -> type[t.Any]:
    """Get adapter class with optimized static mappings and caching.

    This function uses pre-computed static mappings for better performance
    and falls back to dynamic import for custom adapters.
    """
    # Try static mapping first for maximum performance
    mapping_key = f"{category}.{name}" if name != category else category
    if mapping_key in STATIC_ADAPTER_MAPPINGS:
        module_path, class_name = STATIC_ADAPTER_MAPPINGS[mapping_key]
        return _import_adapter_class_cached(module_path, class_name)

    # Fallback to dynamic import for custom adapters
    module_path = f"acb.adapters.{category}.{name}"
    class_name = camelize(category)
    return _import_adapter_class_cached(module_path, class_name)


@lru_cache(maxsize=512)
def _import_adapter_class_cached(module_path: str, class_name: str) -> type:
    """Cached adapter class import to avoid repeated module loading."""
    try:
        module = import_module(module_path)
        adapter_class = getattr(module, class_name)

        # Attach metadata if available
        if hasattr(module, "MODULE_METADATA"):
            metadata = module.MODULE_METADATA
            if isinstance(metadata, AdapterMetadata):
                adapter_class.__module_metadata__ = metadata
            elif isinstance(metadata, dict):
                adapter_class.__module_metadata__ = AdapterMetadata(**metadata)

        return adapter_class  # type: ignore[no-any-return]
    except (ImportError, AttributeError) as e:
        msg = f"Failed to import {module_path}.{class_name}: {e}"
        raise StaticImportError(msg)


def try_import_adapter(category: str, name: str | None = None) -> type[t.Any] | None:
    if name is None:
        adapter_info = get_adapter(category)
        if adapter_info:
            name = adapter_info.name
        else:
            return None
    try:
        return get_adapter_class(category, name)
    except StaticImportError:
        return None


def import_adapter_fast(category: str, name: str | None = None) -> type[t.Any]:
    adapter_class = try_import_adapter(category, name)
    if adapter_class is None:
        available_adapters = list(_get_available_adapters(category))
        msg = (
            f"Could not import {category} adapter '{name}'. "
            f"Available: {available_adapters}"
        )
        raise ImportError(
            msg,
        )
    return adapter_class


def _get_available_adapters(category: str) -> t.Iterator[str]:
    for adapter in adapter_registry.get():
        if adapter.category == category:
            yield adapter.name


_adapter_config_loaded: bool = False


async def _ensure_adapter_configuration() -> None:
    global _adapter_config_loaded
    if _adapter_config_loaded or _should_skip_configuration():
        return
    _adapter_config_loaded = True
    try:
        await _setup_adapter_configuration()
    except Exception:
        import traceback

        traceback.print_exc()


def _should_skip_configuration() -> bool:
    from acb.config import _library_usage_mode

    return _testing or "pytest" in sys.modules or _library_usage_mode


async def _setup_adapter_configuration() -> None:
    cwd_path = AsyncPath(Path.cwd())
    if await _should_skip_project_setup(cwd_path):
        return
    cwd_settings_path = cwd_path / "settings"
    cwd_adapter_settings_path = cwd_settings_path / "adapters.yaml"
    await _create_adapter_settings_if_needed(
        cwd_settings_path,
        cwd_adapter_settings_path,
    )
    await _load_and_apply_adapter_settings(cwd_adapter_settings_path)


async def _should_skip_project_setup(cwd_path: AsyncPath) -> bool:
    if cwd_path.stem == "acb":
        return True
    has_pyproject = await (cwd_path / "pyproject.toml").exists()
    if has_pyproject:
        has_src = await (cwd_path / "src").exists()
        has_acb_dir = await (cwd_path / "acb").exists()
        is_acb_package = (
            has_acb_dir and await (cwd_path / "acb" / "__init__.py").exists()
        )
        has_settings = await (cwd_path / "settings").exists()
        if has_settings:
            return False
        has_package_dir = await (cwd_path / cwd_path.stem).exists()
        return has_src or has_package_dir or is_acb_package

    return False


async def _create_adapter_settings_if_needed(
    cwd_settings_path: AsyncPath,
    cwd_adapter_settings_path: AsyncPath,
) -> None:
    if _testing or "pytest" in sys.modules:
        return

    from acb.config import _library_usage_mode

    if _library_usage_mode:
        return

    if not await cwd_adapter_settings_path.exists() and not _deployed:
        await cwd_settings_path.mkdir(exist_ok=True)
        categories = {a.category for a in adapter_registry.get()}
        settings_dict = {
            cat: None for cat in categories if cat not in ("logger", "config")
        }
        await cwd_adapter_settings_path.write_bytes(
            yaml_encode(settings_dict, sort_keys=True),
        )


async def _load_and_apply_adapter_settings(
    cwd_adapter_settings_path: AsyncPath,
) -> None:
    if not await cwd_adapter_settings_path.exists():
        return

    enabled_adapters = await _load_enabled_adapters(cwd_adapter_settings_path)
    await _ensure_package_registration()
    all_adapters = await _collect_all_adapters()
    _enable_configured_adapters(all_adapters, enabled_adapters)
    _update_adapter_caches()


async def _load_enabled_adapters(
    cwd_adapter_settings_path: AsyncPath,
) -> dict[str, str]:
    _adapters = yaml_decode(await cwd_adapter_settings_path.read_text())
    return {a: m for a, m in _adapters.items() if m}


async def _ensure_package_registration() -> None:
    from acb import ensure_registration_async, get_context

    await ensure_registration_async()
    for pkg in get_context().pkg_registry.get():
        if not pkg.adapters:
            from contextlib import suppress

            with suppress(Exception):
                await register_adapters(pkg.path)


async def _collect_all_adapters() -> list[Adapter]:
    from acb import get_context

    all_adapters = list(adapter_registry.get())
    for pkg in get_context().pkg_registry.get():
        for adapter in pkg.adapters:
            if adapter not in all_adapters:
                adapter_registry.get().append(adapter)
                all_adapters.append(adapter)

    return all_adapters


def _enable_configured_adapters(
    all_adapters: list[Adapter],
    enabled_adapters: dict[str, str],
) -> None:
    for adapter in all_adapters:
        if enabled_adapters.get(adapter.category) == adapter.name:
            adapter.enabled = True


async def _import_adapter_module_for_deps(adapter_category: str) -> None:
    from contextlib import suppress

    if adapter_category == "app":
        for adapter in adapter_registry.get():
            if adapter.category == adapter_category:
                with suppress(ImportError, AdapterNotFound, AdapterNotInstalled):
                    import_module(adapter.module)
                break


async def _find_adapter(adapter_category: str) -> Adapter:
    try:
        adapter = get_adapter(adapter_category)
        if adapter is not None:
            return adapter
        from contextlib import suppress

        current_path = AsyncPath(Path.cwd())
        for pkg_path in (current_path, current_path.parent):
            adapters_dir = pkg_path / "adapters"
            if await adapters_dir.exists():
                with suppress(Exception):
                    await register_adapters(pkg_path)
                    await _ensure_adapter_configuration()
                    adapter = get_adapter(adapter_category)
                    if adapter is not None:
                        return adapter
        msg = f"{adapter_category} adapter not found – check adapters.yaml and ensure package registration"
        raise AdapterNotFound(
            msg,
        )
    except AttributeError:
        msg = f"{adapter_category} adapter not found – check adapters.yaml"
        raise AdapterNotFound(
            msg,
        )


async def _load_module(adapter: Adapter) -> t.Any:
    try:
        module = import_module(adapter.module)
    except ModuleNotFoundError:
        spec = util.spec_from_file_location(adapter.path.stem, str(adapter.path))
        if spec is None:
            msg = f"Failed to create module spec for {adapter.module}"
            raise AdapterNotFound(msg)
        module = util.module_from_spec(spec)
        if spec.loader is not None:
            spec.loader.exec_module(module)
        sys.modules[adapter.name] = module
    return module


async def _initialize_adapter(
    adapter: Adapter,
    module: t.Any,
    adapter_category: str,
) -> t.Any:
    from contextlib import suppress

    adapter_class: t.Any = getattr(module, adapter.class_name)
    adapter_settings = None
    adapter_settings_class_name = f"{adapter.class_name}Settings"
    if hasattr(module, adapter_settings_class_name):
        adapter_settings_class = getattr(module, adapter_settings_class_name)
        with suppress(Exception):
            adapter_settings = adapter_settings_class()
    if adapter_settings is not None:
        from acb.config import Config

        config = await depends.get(Config)
        setattr(config, adapter_category, adapter_settings)
    instance = await depends.get(adapter_class)
    if hasattr(instance, "init"):
        init_result = instance.init()
        if hasattr(init_result, "__await__"):
            await init_result
    from acb.logger import Logger

    logger = await depends.get(Logger)
    logger.info(f"Adapter initialized: {adapter.class_name}")
    return adapter_class


async def _import_adapter(adapter_category: str) -> t.Any:
    if _testing or "pytest" in sys.modules:
        from unittest.mock import MagicMock

        return MagicMock()
    await _ensure_adapter_configuration()
    await _import_adapter_module_for_deps(adapter_category)
    adapter = await _find_adapter(adapter_category)
    module = await _load_module(adapter)
    adapter_class: t.Any = getattr(module, adapter.class_name)
    if adapter.installed:
        return adapter_class
    locks = _adapter_import_locks.get()
    if adapter_category not in locks:
        locks[adapter_category] = asyncio.Lock()
    async with locks[adapter_category]:
        if adapter.installed:
            return adapter_class
        try:
            adapter.installed = True
            return await _initialize_adapter(adapter, module, adapter_category)
        except Exception as e:
            adapter.installed = False
            msg = f"Failed to install {adapter.class_name} adapter: {e}"
            raise AdapterNotInstalled(
                msg,
            )


async def gather_imports(adapter_categories: list[str]) -> t.Any:
    imports = [_import_adapter(category) for category in adapter_categories]
    _imports = await asyncio.gather(*imports, return_exceptions=True)
    results = []
    for result in _imports:
        if isinstance(result, Exception):
            raise result
        results.append(result)
    return results


def import_adapter(adapter_categories: str | list[str] | None = None) -> t.Any:
    if isinstance(adapter_categories, str):
        adapter_info = get_adapter(adapter_categories)
        adapter_name = adapter_info.name if adapter_info else None
        adapter_class = try_import_adapter(adapter_categories, adapter_name)
        if adapter_class:
            return adapter_class
        return import_adapter_with_context([adapter_categories])

    return import_adapter_with_context(adapter_categories)


def _extract_adapter_categories_from_stack() -> list[str]:
    frame_idx = 1
    for i in range(1, min(4, len(stack()))):
        frame_context = str(stack()[i][4]) if stack()[i][4] else ""
        if "import_adapter_with_context" in frame_context:
            frame_idx = i + 1
            break
    context = stack()[frame_idx][4]
    return [
        c.strip()
        for c in (context[0] if context else "")
        .split("=")[0]
        .strip()
        .lower()
        .split(",")
    ]


def _handle_testing_mode(adapter_categories: str | list[str] | None) -> t.Any:
    from unittest.mock import MagicMock

    if isinstance(adapter_categories, str):
        return MagicMock()
    if not adapter_categories:
        try:
            adapter_categories = _extract_adapter_categories_from_stack()
        except (IndexError, AttributeError, TypeError):
            return MagicMock()

    return tuple(MagicMock() for _ in adapter_categories)


def _normalize_adapter_categories(
    adapter_categories: str | list[str] | None,
) -> list[str]:
    if isinstance(adapter_categories, str):
        return [adapter_categories]

    if not adapter_categories:
        try:
            return _extract_adapter_categories_from_stack()
        except (IndexError, AttributeError, TypeError):
            msg = "Could not determine adapter categories from calling context"
            raise ValueError(
                msg,
            )

    return adapter_categories


def import_adapter_with_context(
    adapter_categories: str | list[str] | None = None,
) -> t.Any:
    if _testing or "pytest" in sys.modules:
        return _handle_testing_mode(adapter_categories)

    normalized_categories = _normalize_adapter_categories(adapter_categories)

    try:
        # Check if we're in an async context
        import asyncio

        try:
            asyncio.get_running_loop()
            # We're in an async context - this function shouldn't be called here
            msg = (
                f"import_adapter_with_context() cannot be called from async context. "
                f"Use 'await gather_imports({normalized_categories})' directly."
            )
            raise RuntimeError(msg)
        except RuntimeError as e:
            if "no running event loop" in str(e):
                # No event loop - we need to start one for adapter imports
                # This should only happen in application initialization
                imported_adapters = asyncio.run(gather_imports(normalized_categories))
            else:
                raise
    except Exception as e:
        msg = f"Failed to install adapters {normalized_categories}: {e}"
        raise AdapterNotInstalled(
            msg,
        )
    return (
        imported_adapters[0]
        if len(imported_adapters) == 1
        else tuple(imported_adapters)
    )


async def path_adapters(path: AsyncPath) -> dict[str, list[AsyncPath]]:
    return {
        a.stem: [
            m
            async for m in a.iterdir()
            if not m.name.startswith("_") and m.suffix == ".py"
        ]
        async for a in path.iterdir()
        if await a.is_dir() and (not a.name.startswith("__"))
    }


def create_adapter(
    path: AsyncPath,
    name: str | None = None,
    module: str | None = None,
    class_name: str | None = None,
    category: str | None = None,
    pkg: str | None = None,
    enabled: bool = False,
    installed: bool = False,
) -> Adapter:
    return Adapter(
        path=path,
        name=name or path.stem,
        class_name=class_name or camelize(path.parent.stem),
        category=category or path.parent.stem,
        module=module or ".".join(path.parts[-4:]).removesuffix(".py"),
        pkg=pkg or path.parent.parent.parent.stem,
        enabled=enabled,
        installed=installed,
    )


def extract_adapter_modules(modules: list[AsyncPath], adapter: str) -> list[Adapter]:
    adapter_modules = [create_adapter(p) for p in modules]
    return [a for a in adapter_modules if a.category == adapter]


async def register_adapters(path: AsyncPath) -> list[Adapter]:
    if _testing or "pytest" in sys.modules:
        return []
    _adapters_path = path / "adapters"
    if not (await _adapters_path.exists()):
        return []
    pkg_adapters = await path_adapters(_adapters_path)
    _adapters: list[Adapter] = []
    for adapter_name, modules in pkg_adapters.items():
        _modules = extract_adapter_modules(modules, adapter_name)
        _adapters.extend(_modules)
    registry = adapter_registry.get()
    for a in _adapters:
        module = ".".join(a.module.split(".")[-2:])
        remove = next(
            (_a for _a in registry if ".".join(_a.module.split(".")[-2:]) == module),
            None,
        )
        if remove:
            registry.remove(remove)
        registry.append(a)
        _adapter_import_locks.get()[a.category] = asyncio.Lock()
    _update_adapter_caches()
    return _adapters


adapters = asyncio.run(register_adapters(AsyncPath(__file__).parent.parent))



"""Graph database adapter base classes."""

import typing as t
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum

from pydantic import BaseModel, Field, SecretStr
from acb.cleanup import CleanupMixin
from acb.config import Config, Settings
from acb.depends import Inject, depends
from acb.ssl_config import SSLConfigMixin


class GraphQueryLanguage(str, Enum):
    """Supported graph query languages."""

    CYPHER = "cypher"
    GREMLIN = "gremlin"
    AQL = "aql"
    SPARQL = "sparql"
    PGQ = "pgq"


class GraphTraversalDirection(str, Enum):
    """Graph traversal direction options."""

    OUT = "out"
    IN = "in"
    BOTH = "both"


class GraphNodeModel(BaseModel):
    """Base model for graph nodes."""

    id: str | None = None
    labels: list[str] = Field(default_factory=list)
    properties: dict[str, t.Any] = Field(default_factory=dict)
    created_at: datetime | None = None
    updated_at: datetime | None = None


class GraphEdgeModel(BaseModel):
    """Base model for graph edges/relationships."""

    id: str | None = None
    type: str
    from_node: str
    to_node: str
    properties: dict[str, t.Any] = Field(default_factory=dict)
    created_at: datetime | None = None
    updated_at: datetime | None = None


class GraphPathModel(BaseModel):
    """Base model for graph paths."""

    nodes: list[GraphNodeModel]
    edges: list[GraphEdgeModel]
    length: int
    weight: float | None = None


class GraphSchemaModel(BaseModel):
    """Base model for graph schema information."""

    node_types: list[str] = Field(default_factory=list)
    edge_types: list[str] = Field(default_factory=list)
    constraints: list[dict[str, t.Any]] = Field(default_factory=list)
    indexes: list[dict[str, t.Any]] = Field(default_factory=list)


class GraphQueryResult(BaseModel):
    """Base model for graph query results."""

    nodes: list[GraphNodeModel] = Field(default_factory=list)
    edges: list[GraphEdgeModel] = Field(default_factory=list)
    paths: list[GraphPathModel] = Field(default_factory=list)
    records: list[dict[str, t.Any]] = Field(default_factory=list)
    metadata: dict[str, t.Any] = Field(default_factory=dict)
    execution_time: float | None = None
    query_language: GraphQueryLanguage | None = None


class GraphBaseSettings(Settings, SSLConfigMixin):
    """Base settings for graph database adapters.

    Note: SSL configuration fields are defined for documentation purposes but are
    excluded from Pydantic validation to avoid conflicts with SSLConfigMixin properties.
    These fields are handled manually in __init__.
    """

    # Pydantic config to exclude SSL fields from validation (handled in __init__)
    model_config = {
        "extra": "allow",  # Allow extra fields for SSL config
    }

    # Connection settings
    host: SecretStr = SecretStr("127.0.0.1")
    port: int | None = None
    user: SecretStr | None = None
    password: SecretStr | None = None
    database: str | None = None
    auth_token: SecretStr | None = None

    # Connection pooling
    max_connections: int = 50
    min_connections: int = 1
    connection_timeout: float = 30.0
    idle_timeout: float = 600.0
    max_connection_lifetime: float = 3600.0

    # Query settings
    query_timeout: float = 300.0
    max_query_size: int = 1024 * 1024  # 1MB
    enable_query_logging: bool = False

    # Transaction settings
    transaction_timeout: float = 600.0
    auto_commit: bool = True

    # Performance settings
    enable_caching: bool = True
    cache_ttl: int = 300
    batch_size: int = 1000
    parallel_queries: bool = False

    # SSL/TLS Configuration (handled manually in __init__, not validated by Pydantic)
    # These are not Pydantic fields - they are popped in __init__ before validation
    # ssl_enabled: bool = False
    # ssl_cert_path: str | None = None
    # ssl_key_path: str | None = None
    # ssl_ca_path: str | None = None
    # ssl_verify_mode: str = "required"
    # tls_version: str = "TLSv1.2"

    @depends.inject
    def __init__(self, config: Inject[Config], **values: t.Any) -> None:
        # Extract SSL configuration parameters
        ssl_enabled = values.pop("ssl_enabled", False)
        ssl_cert_path = values.pop("ssl_cert_path", None)
        ssl_key_path = values.pop("ssl_key_path", None)
        ssl_ca_path = values.pop("ssl_ca_path", None)
        ssl_verify_mode = values.pop("ssl_verify_mode", "required")
        values.pop("tls_version", "TLSv1.2")

        super().__init__(**values)
        SSLConfigMixin.__init__(self)

        # Configure SSL if enabled
        if ssl_enabled:
            from acb.ssl_config import SSLVerifyMode

            verify_mode_map = {
                "none": SSLVerifyMode.NONE,
                "optional": SSLVerifyMode.OPTIONAL,
                "required": SSLVerifyMode.REQUIRED,
            }
            verify_mode = verify_mode_map.get(ssl_verify_mode, SSLVerifyMode.REQUIRED)

            self.configure_ssl(
                enabled=True,
                cert_path=ssl_cert_path,
                key_path=ssl_key_path,
                ca_path=ssl_ca_path,
                verify_mode=verify_mode,
                check_hostname=verify_mode == SSLVerifyMode.REQUIRED,
            )


class GraphProtocol(t.Protocol):
    """Protocol for graph database operations."""

    async def execute_query(
        self,
        query: str,
        parameters: dict[str, t.Any] | None = None,
    ) -> GraphQueryResult: ...

    async def create_node(
        self,
        labels: list[str] | None = None,
        properties: dict[str, t.Any] | None = None,
    ) -> GraphNodeModel: ...

    async def create_edge(
        self,
        edge_type: str,
        from_node_id: str,
        to_node_id: str,
        properties: dict[str, t.Any] | None = None,
    ) -> GraphEdgeModel: ...

    async def get_node(self, node_id: str) -> GraphNodeModel | None: ...

    async def get_edge(self, edge_id: str) -> GraphEdgeModel | None: ...

    async def find_path(
        self,
        from_node_id: str,
        to_node_id: str,
        max_depth: int | None = None,
        direction: GraphTraversalDirection = GraphTraversalDirection.BOTH,
    ) -> list[GraphPathModel]: ...


class GraphBase(CleanupMixin, ABC):
    """Base class for graph database adapters."""

    config: Inject[Config]
    logger: Inject[t.Any]

    def __init__(self, **kwargs: t.Any) -> None:
        CleanupMixin.__init__(self)
        self._client: t.Any = None
        self._client_lock: t.Any | None = (
            None  # Type as Any | None to satisfy async Lock
        )
        self._transaction: t.Any = None

    @property
    @abstractmethod
    def query_language(self) -> GraphQueryLanguage:
        """Return the primary query language supported by this adapter."""

    @property
    @abstractmethod
    def supported_features(self) -> list[str]:
        """Return list of supported features for this adapter."""

    async def _ensure_client(self) -> t.Any:
        """Ensure client connection with lazy initialization."""
        if self._client is None:
            if self._client_lock is None:
                import asyncio

                self._client_lock = asyncio.Lock()
            async with self._client_lock:
                if self._client is None:
                    self._client = await self._create_client()
        return self._client

    @abstractmethod
    async def _create_client(self) -> t.Any:
        """Create the database client connection."""

    async def _cleanup_resources(self) -> None:
        """Clean up graph-specific resources."""
        if self._transaction is not None:
            await self._rollback_transaction()
            self._transaction = None

        if self._client is not None:
            self.register_resource(self._client)
            self._client = None

        await super().cleanup()

    # Transaction Management
    async def begin_transaction(self) -> None:
        """Begin a new transaction."""
        client = await self._ensure_client()
        self._transaction = await self._begin_transaction(client)

    @abstractmethod
    async def _begin_transaction(self, client: t.Any) -> t.Any:
        """Implementation-specific transaction begin."""

    async def commit_transaction(self) -> None:
        """Commit the current transaction."""
        if self._transaction is None:
            msg = "No active transaction to commit"
            raise RuntimeError(msg)
        await self._commit_transaction(self._transaction)
        self._transaction = None

    @abstractmethod
    async def _commit_transaction(self, transaction: t.Any) -> None:
        """Implementation-specific transaction commit."""

    async def rollback_transaction(self) -> None:
        """Rollback the current transaction."""
        if self._transaction is None:
            msg = "No active transaction to rollback"
            raise RuntimeError(msg)
        await self._rollback_transaction()
        self._transaction = None

    @abstractmethod
    async def _rollback_transaction(self) -> None:
        """Implementation-specific transaction rollback."""

    # Core Graph Operations
    async def execute_query(
        self,
        query: str,
        parameters: dict[str, t.Any] | None = None,
        timeout: float | None = None,
    ) -> GraphQueryResult:
        """Execute a graph query and return results."""
        return await self._execute_query(query, parameters, timeout)

    @abstractmethod
    async def _execute_query(
        self,
        query: str,
        parameters: dict[str, t.Any] | None = None,
        timeout: float | None = None,
    ) -> GraphQueryResult:
        """Implementation-specific query execution."""

    async def execute_batch_queries(
        self,
        queries: list[tuple[str, dict[str, t.Any] | None]],
    ) -> list[GraphQueryResult]:
        """Execute multiple queries in batch."""
        results = []
        for query, params in queries:
            result = await self.execute_query(query, params)
            results.append(result)
        return results

    # Node Operations
    async def create_node(
        self,
        labels: list[str] | None = None,
        properties: dict[str, t.Any] | None = None,
    ) -> GraphNodeModel:
        """Create a new node in the graph."""
        return await self._create_node(labels or [], properties or {})

    @abstractmethod
    async def _create_node(
        self,
        labels: list[str],
        properties: dict[str, t.Any],
    ) -> GraphNodeModel:
        """Implementation-specific node creation."""

    async def get_node(self, node_id: str) -> GraphNodeModel | None:
        """Get a node by ID."""
        return await self._get_node(node_id)

    @abstractmethod
    async def _get_node(self, node_id: str) -> GraphNodeModel | None:
        """Implementation-specific node retrieval."""

    async def update_node(
        self,
        node_id: str,
        properties: dict[str, t.Any],
    ) -> GraphNodeModel:
        """Update node properties."""
        return await self._update_node(node_id, properties)

    @abstractmethod
    async def _update_node(
        self,
        node_id: str,
        properties: dict[str, t.Any],
    ) -> GraphNodeModel:
        """Implementation-specific node update."""

    async def delete_node(self, node_id: str) -> bool:
        """Delete a node from the graph."""
        return await self._delete_node(node_id)

    @abstractmethod
    async def _delete_node(self, node_id: str) -> bool:
        """Implementation-specific node deletion."""

    # Edge Operations
    async def create_edge(
        self,
        edge_type: str,
        from_node_id: str,
        to_node_id: str,
        properties: dict[str, t.Any] | None = None,
    ) -> GraphEdgeModel:
        """Create a new edge between nodes."""
        return await self._create_edge(
            edge_type,
            from_node_id,
            to_node_id,
            properties or {},
        )

    @abstractmethod
    async def _create_edge(
        self,
        edge_type: str,
        from_node_id: str,
        to_node_id: str,
        properties: dict[str, t.Any],
    ) -> GraphEdgeModel:
        """Implementation-specific edge creation."""

    async def get_edge(self, edge_id: str) -> GraphEdgeModel | None:
        """Get an edge by ID."""
        return await self._get_edge(edge_id)

    @abstractmethod
    async def _get_edge(self, edge_id: str) -> GraphEdgeModel | None:
        """Implementation-specific edge retrieval."""

    async def update_edge(
        self,
        edge_id: str,
        properties: dict[str, t.Any],
    ) -> GraphEdgeModel:
        """Update edge properties."""
        return await self._update_edge(edge_id, properties)

    @abstractmethod
    async def _update_edge(
        self,
        edge_id: str,
        properties: dict[str, t.Any],
    ) -> GraphEdgeModel:
        """Implementation-specific edge update."""

    async def delete_edge(self, edge_id: str) -> bool:
        """Delete an edge from the graph."""
        return await self._delete_edge(edge_id)

    @abstractmethod
    async def _delete_edge(self, edge_id: str) -> bool:
        """Implementation-specific edge deletion."""

    # Graph Traversal Operations
    async def find_path(
        self,
        from_node_id: str,
        to_node_id: str,
        max_depth: int | None = None,
        direction: GraphTraversalDirection = GraphTraversalDirection.BOTH,
    ) -> list[GraphPathModel]:
        """Find paths between two nodes."""
        return await self._find_path(from_node_id, to_node_id, max_depth, direction)

    @abstractmethod
    async def _find_path(
        self,
        from_node_id: str,
        to_node_id: str,
        max_depth: int | None,
        direction: GraphTraversalDirection,
    ) -> list[GraphPathModel]:
        """Implementation-specific path finding."""

    async def find_shortest_path(
        self,
        from_node_id: str,
        to_node_id: str,
        weight_property: str | None = None,
    ) -> GraphPathModel | None:
        """Find the shortest path between two nodes."""
        return await self._find_shortest_path(from_node_id, to_node_id, weight_property)

    @abstractmethod
    async def _find_shortest_path(
        self,
        from_node_id: str,
        to_node_id: str,
        weight_property: str | None,
    ) -> GraphPathModel | None:
        """Implementation-specific shortest path finding."""

    async def get_neighbors(
        self,
        node_id: str,
        direction: GraphTraversalDirection = GraphTraversalDirection.BOTH,
        edge_types: list[str] | None = None,
    ) -> list[GraphNodeModel]:
        """Get neighboring nodes."""
        return await self._get_neighbors(node_id, direction, edge_types)

    @abstractmethod
    async def _get_neighbors(
        self,
        node_id: str,
        direction: GraphTraversalDirection,
        edge_types: list[str] | None,
    ) -> list[GraphNodeModel]:
        """Implementation-specific neighbor retrieval."""

    # Schema and Analytics Operations
    async def get_schema(self) -> GraphSchemaModel:
        """Get graph schema information."""
        return await self._get_schema()

    @abstractmethod
    async def _get_schema(self) -> GraphSchemaModel:
        """Implementation-specific schema retrieval."""

    async def create_index(
        self,
        labels: list[str],
        properties: list[str],
        index_type: str = "btree",
    ) -> bool:
        """Create an index on specified properties."""
        return await self._create_index(labels, properties, index_type)

    @abstractmethod
    async def _create_index(
        self,
        labels: list[str],
        properties: list[str],
        index_type: str,
    ) -> bool:
        """Implementation-specific index creation."""

    async def drop_index(self, index_name: str) -> bool:
        """Drop an index."""
        return await self._drop_index(index_name)

    @abstractmethod
    async def _drop_index(self, index_name: str) -> bool:
        """Implementation-specific index deletion."""

    # Bulk Operations
    async def bulk_create_nodes(
        self,
        nodes: list[dict[str, t.Any]],
    ) -> list[GraphNodeModel]:
        """Create multiple nodes in bulk."""
        return await self._bulk_create_nodes(nodes)

    @abstractmethod
    async def _bulk_create_nodes(
        self,
        nodes: list[dict[str, t.Any]],
    ) -> list[GraphNodeModel]:
        """Implementation-specific bulk node creation."""

    async def bulk_create_edges(
        self,
        edges: list[dict[str, t.Any]],
    ) -> list[GraphEdgeModel]:
        """Create multiple edges in bulk."""
        return await self._bulk_create_edges(edges)

    @abstractmethod
    async def _bulk_create_edges(
        self,
        edges: list[dict[str, t.Any]],
    ) -> list[GraphEdgeModel]:
        """Implementation-specific bulk edge creation."""

    # Utility Methods
    async def count_nodes(self, labels: list[str] | None = None) -> int:
        """Count nodes in the graph."""
        return await self._count_nodes(labels)

    @abstractmethod
    async def _count_nodes(self, labels: list[str] | None) -> int:
        """Implementation-specific node counting."""

    async def count_edges(self, edge_types: list[str] | None = None) -> int:
        """Count edges in the graph."""
        return await self._count_edges(edge_types)

    @abstractmethod
    async def _count_edges(self, edge_types: list[str] | None) -> int:
        """Implementation-specific edge counting."""

    async def clear_graph(self) -> bool:
        """Clear all nodes and edges from the graph."""
        return await self._clear_graph()

    @abstractmethod
    async def _clear_graph(self) -> bool:
        """Implementation-specific graph clearing."""



"""Tests for the ACB MCP registry module."""
import pytest
from unittest.mock import Mock, MagicMock, AsyncMock, patch

from acb.mcp.registry import ComponentRegistry


class TestComponentRegistry:
    """Test the ComponentRegistry class."""

    def test_initialization(self) -> None:
        """Test basic initialization of ComponentRegistry."""
        registry = ComponentRegistry()

        assert registry._actions == {}
        assert registry._adapters == {}
        assert registry._services == {}
        assert registry._events == {}
        assert registry._initialized is False
        assert hasattr(registry, 'config')
        assert hasattr(registry, 'logger')

    @pytest.mark.asyncio
    async def test_initialize_first_time(self) -> None:
        """Test initializing the registry for the first time."""
        registry = ComponentRegistry()

        # Mock the dependencies to avoid actual dependency injection
        with (
            patch.object(registry, 'logger') as mock_logger,
            patch('acb.mcp.registry.import_adapter'),
            patch('acb.mcp.registry.depends') as mock_depends,
        ):
            # Mock the actions imports to work without actual modules
            mock_compress = Mock()
            mock_encode = Mock()
            mock_hash = Mock()

            with (
                patch.dict('sys.modules', {
                    'acb.actions.compress': mock_compress,
                    'acb.actions.encode': mock_encode,
                    'acb.actions.hash': mock_hash,
                }),
                patch('builtins.__import__', side_effect=[
                    Mock(),  # For acb.actions
                    mock_compress,  # For acb.actions.compress
                    Mock(),  # For acb.actions
                    mock_encode,  # For acb.actions.encode
                    Mock(),  # For acb.actions
                    mock_hash,  # For acb.actions.hash
                ])
            ):
                await registry.initialize()

                assert registry._initialized is True
                assert 'compress' in registry._actions
                assert 'encode' in registry._actions
                assert 'hash' in registry._actions
                mock_logger.info.assert_called()

    @pytest.mark.asyncio
    async def test_initialize_already_initialized(self) -> None:
        """Test that initialize is skipped when already initialized."""
        registry = ComponentRegistry()
        registry._initialized = True

        with patch.object(registry, 'logger') as mock_logger:
            await registry.initialize()
            # The logger should not have been called since it was already initialized
            mock_logger.info.assert_not_called()

    @pytest.mark.asyncio
    async def test_register_builtin_actions_success(self) -> None:
        """Test registering builtin actions successfully."""
        registry = ComponentRegistry()

        # Mock action modules
        mock_compress = Mock()
        mock_encode = Mock()
        mock_hash = Mock()

        with (
            patch('sys.modules', {}),  # Clear the sys.modules cache
            patch('builtins.__import__', side_effect=lambda name, *args, **kwargs: {
                'acb.actions.compress': mock_compress,
                'acb.actions.encode': mock_encode,
                'acb.actions.hash': mock_hash,
            }.get(name, Mock())),
            patch.object(registry, 'logger') as mock_logger,
        ):
            await registry._register_builtin_actions()

            assert 'compress' in registry._actions
            assert 'encode' in registry._actions
            assert 'hash' in registry._actions
            mock_logger.debug.assert_called()

    @pytest.mark.asyncio
    async def test_register_builtin_actions_failure(self) -> None:
        """Test handling failures during action registration."""
        registry = ComponentRegistry()

        with (
            patch('builtins.__import__', side_effect=ImportError("Module not found")),
            patch.object(registry, 'logger') as mock_logger,
        ):
            await registry._register_builtin_actions()
            mock_logger.warning.assert_called()

    @pytest.mark.asyncio
    async def test_register_configured_adapters(self) -> None:
        """Test registering configured adapters."""
        registry = ComponentRegistry()

        # Mock config to have adapter configurations
        mock_config = Mock()
        mock_config.adapters = {"cache": "memory", "storage": "file"}
        registry.config = mock_config

        # Mock adapter imports and dependencies
        mock_cache_adapter = Mock()
        mock_storage_adapter = Mock()

        with (
            patch('acb.mcp.registry.import_adapter', side_effect=[
                mock_cache_adapter, mock_storage_adapter
            ]),
            patch('acb.mcp.registry.depends') as mock_depends,
            patch.object(registry, 'logger') as mock_logger,
        ):
            mock_depends.get.side_effect = [mock_cache_adapter, mock_storage_adapter]

            await registry._register_configured_adapters()

            assert "cache" in registry._adapters
            assert "storage" in registry._adapters
            mock_logger.debug.assert_called()

    @pytest.mark.asyncio
    async def test_register_configured_adapters_with_error(self) -> None:
        """Test handling errors when registering adapters."""
        registry = ComponentRegistry()

        # Mock config to have adapter configurations
        mock_config = Mock()
        mock_config.adapters = {"invalid_adapter": "invalid"}
        registry.config = mock_config

        with (
            patch('acb.mcp.registry.import_adapter', side_effect=ImportError("Not found")),
            patch.object(registry, 'logger') as mock_logger,
        ):
            await registry._register_configured_adapters()
            mock_logger.warning.assert_called()

    @pytest.mark.asyncio
    async def test_register_services_placeholder(self) -> None:
        """Test the service registration placeholder."""
        registry = ComponentRegistry()

        with patch.object(registry, 'logger') as mock_logger:
            await registry._register_services()
            mock_logger.debug.assert_called_with("Services registration placeholder")

    def test_get_actions(self) -> None:
        """Test getting all actions."""
        registry = ComponentRegistry()
        registry._actions = {"test": Mock()}

        actions = registry.get_actions()
        assert "test" in actions
        # Should be a copy, not the original dict
        assert actions is not registry._actions

    def test_get_action(self) -> None:
        """Test getting a specific action."""
        registry = ComponentRegistry()
        registry._actions = {"test": "value"}

        action = registry.get_action("test")
        assert action == "value"

        missing_action = registry.get_action("missing")
        assert missing_action is None

    def test_get_adapters(self) -> None:
        """Test getting all adapters."""
        registry = ComponentRegistry()
        registry._adapters = {"test": Mock()}

        adapters = registry.get_adapters()
        assert "test" in adapters
        # Should be a copy, not the original dict
        assert adapters is not registry._adapters

    def test_get_adapter(self) -> None:
        """Test getting a specific adapter."""
        registry = ComponentRegistry()
        registry._adapters = {"test": "value"}

        adapter = registry.get_adapter("test")
        assert adapter == "value"

        missing_adapter = registry.get_adapter("missing")
        assert missing_adapter is None

    def test_get_services(self) -> None:
        """Test getting all services."""
        registry = ComponentRegistry()
        registry._services = {"test": Mock()}

        services = registry.get_services()
        assert "test" in services
        # Should be a copy, not the original dict
        assert services is not registry._services

    def test_get_service(self) -> None:
        """Test getting a specific service."""
        registry = ComponentRegistry()
        registry._services = {"test": "value"}

        service = registry.get_service("test")
        assert service == "value"

        missing_service = registry.get_service("missing")
        assert missing_service is None

    def test_get_events(self) -> None:
        """Test getting all events."""
        registry = ComponentRegistry()
        registry._events = {"test": Mock()}

        events = registry.get_events()
        assert "test" in events
        # Should be a copy, not the original dict
        assert events is not registry._events

    def test_get_event(self) -> None:
        """Test getting a specific event."""
        registry = ComponentRegistry()
        registry._events = {"test": "value"}

        event = registry.get_event("test")
        assert event == "value"

        missing_event = registry.get_event("missing")
        assert missing_event is None

    @pytest.mark.asyncio
    async def test_cleanup(self) -> None:
        """Test cleaning up the registry."""
        registry = ComponentRegistry()
        registry._actions = {"test": "value"}
        registry._adapters = {"test": "value"}
        registry._services = {"test": "value"}
        registry._events = {"test": "value"}
        registry._initialized = True

        with patch.object(registry, 'logger') as mock_logger:
            await registry.cleanup()

            assert registry._actions == {}
            assert registry._adapters == {}
            assert registry._services == {}
            assert registry._events == {}
            assert registry._initialized is False
            mock_logger.info.assert_called_with("ACB Component Registry cleaned up")



"""ACB Testing Utilities.

Provides utility functions for testing ACB components, adapters,
and services with interface validation and test environment setup.

Features:
- Interface validation for adapters, services, and actions
- Test configuration creation
- Test component creation
- Environment setup and teardown
- Test suite orchestration
"""

import asyncio
import typing as t
from contextlib import suppress
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock

from acb.config import Config
from acb.depends import depends


def assert_adapter_interface(
    adapter_class: type,
    expected_methods: list[str] | None = None,
) -> None:
    """Assert that an adapter implements the expected interface."""
    if expected_methods is None:
        # Common adapter methods
        expected_methods = [
            "_ensure_client",
            "_create_client",
        ]

    for method_name in expected_methods:
        if not hasattr(adapter_class, method_name):
            raise AssertionError(f"Adapter missing method: {method_name}")

        method = getattr(adapter_class, method_name)
        if not callable(method):
            raise AssertionError(
                f"Adapter method {method_name} is not callable",
            )

        # Check if async method
        if method_name.startswith("_") or asyncio.iscoroutinefunction(method):
            if not asyncio.iscoroutinefunction(method):
                raise AssertionError(f"Method {method_name} should be async")


def assert_service_interface(
    service_class: type,
    expected_methods: list[str] | None = None,
) -> None:
    """Assert that a service implements the expected interface."""
    if expected_methods is None:
        # Common service methods vary by type, but check basic structure
        expected_methods = []

    # Check that service has SERVICE_METADATA if it's a discoverable service
    if hasattr(service_class, "SERVICE_METADATA"):
        metadata = service_class.SERVICE_METADATA
        if metadata is None:
            raise AssertionError("Service metadata should not be None")
        if not hasattr(metadata, "name"):
            raise AssertionError("Service metadata should have name")
        if not hasattr(metadata, "category"):
            raise AssertionError("Service metadata should have category")

    for method_name in expected_methods:
        if not hasattr(service_class, method_name):
            raise AssertionError(f"Service missing method: {method_name}")

        method = getattr(service_class, method_name)
        if not callable(method):
            raise AssertionError(
                f"Service method {method_name} is not callable",
            )


def assert_action_interface(
    action_module: t.Any,
    expected_functions: list[str] | None = None,
) -> None:
    """Assert that an action module implements the expected interface."""
    if expected_functions is None:
        # Actions typically have these patterns
        expected_functions = []

    for function_name in expected_functions:
        if not hasattr(action_module, function_name):
            raise AssertionError(f"Action missing function: {function_name}")

        function = getattr(action_module, function_name)
        if not callable(function):
            raise AssertionError(
                f"Action function {function_name} is not callable",
            )


def create_test_config(overrides: dict[str, t.Any] | None = None) -> Config:
    """Create a test configuration with sensible defaults."""
    base_config = {
        "app": {
            "name": "test-app",
            "version": "1.0.0",
            "environment": "test",
        },
        "debug": {
            "enabled": True,
            "level": "DEBUG",
        },
        "adapters": {
            "cache": "memory",
            "storage": "memory",
            "sql": "sqlite",
        },
        "testing": {
            "mock_external_services": True,
            "use_in_memory_db": True,
            "cleanup_resources": True,
        },
    }

    if overrides:
        # Deep merge overrides
        def deep_merge(
            base: dict[str, t.Any], override: dict[str, t.Any]
        ) -> dict[str, t.Any]:
            result = base.copy()
            for key, value in override.items():
                if (
                    key in result
                    and isinstance(result[key], dict)
                    and isinstance(value, dict)
                ):
                    result[key] = deep_merge(result[key], value)
                else:
                    result[key] = value
            return result

        base_config = deep_merge(base_config, overrides)

    config = Config()
    # Store config data as test attribute for test utilities
    config._test_config_data = base_config  # type: ignore[attr-defined]
    return config


async def create_test_adapter(
    adapter_type: str, config: dict[str, t.Any] | None = None
) -> t.Any:
    """Create a test adapter instance with mock configuration."""
    from acb.adapters import import_adapter

    try:
        # Import the adapter class
        AdapterClass = import_adapter(adapter_type)

        # Create instance
        adapter = AdapterClass()

        # Mock the settings if needed
        if config:
            adapter._settings = MagicMock()
            for key, value in config.items():
                setattr(adapter._settings, key, value)

        return adapter

    except Exception:
        # Return a mock adapter if real one not available
        mock_adapter = AsyncMock()

        # Add common adapter methods
        mock_adapter._ensure_client = AsyncMock()
        mock_adapter._create_client = AsyncMock()

        # Configure based on adapter type
        if adapter_type == "cache":
            mock_adapter.get = AsyncMock(return_value=None)
            mock_adapter.set = AsyncMock(return_value=True)
            mock_adapter.delete = AsyncMock(return_value=True)

        elif adapter_type == "storage":
            mock_adapter.read = AsyncMock(return_value=b"test data")
            mock_adapter.write = AsyncMock(return_value=True)
            mock_adapter.exists = AsyncMock(return_value=True)

        elif adapter_type == "sql":
            mock_result = MagicMock()
            mock_result.rowcount = 1
            mock_adapter.execute = AsyncMock(return_value=mock_result)
            mock_adapter.fetch_one = AsyncMock(return_value={"id": 1})
            mock_adapter.fetch_all = AsyncMock(return_value=[{"id": 1}])

        return mock_adapter


async def create_test_service(
    service_type: str, config: dict[str, t.Any] | None = None
) -> t.Any:
    """Create a test service instance with mock configuration."""
    try:
        from acb.services.discovery import import_service

        # Import the service class
        ServiceClass = import_service(service_type)

        # Create instance
        service = ServiceClass()

        # Mock the settings if needed
        if config:
            service._settings = MagicMock()
            for key, value in config.items():
                setattr(service._settings, key, value)

        return service

    except Exception:
        # Return a mock service if real one not available
        mock_service = AsyncMock()

        # Configure based on service type
        if service_type == "performance":
            mock_service.optimize = AsyncMock(return_value={"optimized": True})
            mock_service.get_metrics = AsyncMock(return_value={"cpu": 50})

        elif service_type == "health":
            mock_service.check = AsyncMock(return_value={"status": "healthy"})
            mock_service.get_status = AsyncMock(return_value="healthy")

        elif service_type == "validation":
            mock_service.validate = AsyncMock(
                return_value={"valid": True, "errors": []},
            )
            mock_service.sanitize = AsyncMock(return_value="clean data")

        return mock_service


async def setup_test_environment(
    config: dict[str, t.Any] | None = None,
) -> dict[str, t.Any]:
    """Setup a complete test environment with all necessary components."""
    # Create test config
    test_config = create_test_config(config)

    # Initialize dependency injection
    depends.set(Config, test_config)

    # Create test adapters based on configuration
    adapters: dict[str, t.Any] = {}
    # Access test config data
    adapter_configs = getattr(test_config, "_test_config_data", {}).get("adapters", {})

    for adapter_type in adapter_configs:
        with suppress(Exception):
            adapter = await create_test_adapter(adapter_type)
            adapters[adapter_type] = adapter
            depends.set(type(adapter), adapter)

    # Create test environment info
    # Access depends instances using get() method instead of private attribute
    dependencies: dict[str, t.Any] = {}
    return {
        "config": test_config,
        "adapters": adapters,
        "dependencies": dependencies,
        "status": "ready",
        "created_at": "2024-01-01T12:00:00Z",
    }


async def teardown_test_environment(environment: dict[str, t.Any]) -> None:
    """Teardown a test environment and cleanup resources."""
    # Cleanup adapters
    for adapter in environment.get("adapters", {}).values():
        with suppress(Exception):
            if hasattr(adapter, "cleanup"):
                await adapter.cleanup()
            elif hasattr(adapter, "close"):
                await adapter.close()
            elif hasattr(adapter, "__aexit__"):
                await adapter.__aexit__(None, None, None)

    # Clear dependency injection - note: we skip clearing private instances in testing

    # Update environment status
    environment["status"] = "cleaned_up"


async def run_acb_test_suite(
    test_functions: list[t.Callable[..., t.Any]],
    environment_config: dict[str, t.Any] | None = None,
    parallel: bool = False,
) -> dict[str, t.Any]:
    """Run a complete ACB test suite with environment setup and teardown."""
    environment = await setup_test_environment(environment_config)
    start_time = asyncio.get_event_loop().time()

    try:
        results = (
            await _run_tests_parallel(test_functions)
            if parallel
            else await _run_tests_sequential(test_functions)
        )
    finally:
        await teardown_test_environment(environment)

    execution_time = asyncio.get_event_loop().time() - start_time

    return _build_test_summary(
        test_functions, results, execution_time, parallel, environment_config
    )


async def _run_tests_parallel(
    test_functions: list[t.Callable[..., t.Any]],
) -> list[dict[str, t.Any]]:
    """Run tests in parallel."""
    tasks = [_wrap_test_function(func) for func in test_functions]
    test_results = await asyncio.gather(*tasks, return_exceptions=True)

    return [
        _create_test_result(test_functions[i], result)
        for i, result in enumerate(test_results)
    ]


async def _run_tests_sequential(
    test_functions: list[t.Callable[..., t.Any]],
) -> list[dict[str, t.Any]]:
    """Run tests sequentially."""
    results = []
    for test_func in test_functions:
        try:
            result = await _execute_test_function(test_func)
            results.append(_create_success_result(test_func, result))
        except Exception as e:
            results.append(_create_failure_result(test_func, e))

    return results


async def _wrap_test_function(test_func: t.Callable[..., t.Any]) -> t.Any:
    """Wrap test function for parallel execution."""
    if asyncio.iscoroutinefunction(test_func):
        return await test_func()

    async def async_wrapper() -> t.Any:
        return test_func()

    return await async_wrapper()


async def _execute_test_function(test_func: t.Callable[..., t.Any]) -> t.Any:
    """Execute a test function (async or sync)."""
    if asyncio.iscoroutinefunction(test_func):
        return await test_func()
    return test_func()


def _create_test_result(
    test_func: t.Callable[..., t.Any], result: t.Any
) -> dict[str, t.Any]:
    """Create test result from function and execution result."""
    if isinstance(result, BaseException):
        return _create_failure_result(test_func, result)
    return _create_success_result(test_func, result)


def _create_success_result(
    test_func: t.Callable[..., t.Any], result: t.Any
) -> dict[str, t.Any]:
    """Create success result."""
    return {
        "test": test_func.__name__,
        "status": "passed",
        "result": result,
    }


def _create_failure_result(
    test_func: t.Callable[..., t.Any], error: BaseException
) -> dict[str, t.Any]:
    """Create failure result."""
    return {
        "test": test_func.__name__,
        "status": "failed",
        "error": str(error),
        "error_type": type(error).__name__,
    }


def _build_test_summary(
    test_functions: list[t.Callable[..., t.Any]],
    results: list[dict[str, t.Any]],
    execution_time: float,
    parallel: bool,
    environment_config: dict[str, t.Any] | None,
) -> dict[str, t.Any]:
    """Build test execution summary."""
    passed_tests = [r for r in results if r["status"] == "passed"]

    return {
        "total_tests": len(test_functions),
        "passed": len(passed_tests),
        "failed": len(results) - len(passed_tests),
        "success_rate": len(passed_tests) / len(test_functions)
        if test_functions
        else 1.0,
        "execution_time": execution_time,
        "parallel": parallel,
        "results": results,
        "environment": environment_config,
        "timestamp": "2024-01-01T12:00:00Z",
    }


def validate_test_result(
    result: dict[str, t.Any],
    expected_status: str = "passed",
) -> None:
    """Validate that a test result meets expectations."""
    if "status" not in result:
        raise AssertionError("Test result missing status")
    if result["status"] != expected_status:
        raise AssertionError(
            f"Expected {expected_status}, got {result['status']}",
        )

    if expected_status == "passed":
        if "error" in result and result["error"] is not None:
            raise AssertionError("Passed test should not have errors")
    elif expected_status == "failed":
        if "error" not in result:
            raise AssertionError("Failed test should have error information")


def create_mock_dependency(
    interface_type: type,
    behavior: dict[str, t.Any] | None = None,
) -> MagicMock:
    """Create a mock dependency that implements a specific interface."""
    mock = MagicMock(spec=interface_type)

    # Configure behavior if provided
    if behavior:
        for attr_name, attr_value in behavior.items():
            if callable(attr_value):
                setattr(mock, attr_name, attr_value)
            else:
                mock.configure_mock(**{attr_name: attr_value})

    return mock


def assert_dependency_injected(dependency_type: type) -> t.Any:
    """Assert that a dependency is properly injected."""
    instance = depends.get(dependency_type)
    if instance is None:
        raise AssertionError(
            f"Dependency {dependency_type.__name__} not injected",
        )
    return instance


def create_temporary_config_file(
    config_data: dict[str, t.Any], file_path: Path
) -> Path:
    """Create a temporary configuration file for testing."""
    import yaml

    # Ensure parent directory exists
    file_path.parent.mkdir(parents=True, exist_ok=True)

    # Write config data
    with file_path.open("w") as f:
        yaml.dump(config_data, f, default_flow_style=False)

    return file_path



"""Tests for validation decorators."""

import pytest
from unittest.mock import Mock, AsyncMock, patch

from acb.depends import depends
from acb.services.validation import (
    ValidationError,
    ValidationService,
    validate_input,
    validate_output,
    sanitize_input,
    validate_schema,
    validate_contracts,
)
from acb.services.validation._base import ValidationResult
from acb.services.validation.schemas import StringValidationSchema


class TestValidationDecorators:
    """Tests for validation decorators."""

    @pytest.fixture
    def mock_validation_service(self) -> Mock:
        """Mock ValidationService."""
        service = Mock(spec=ValidationService)
        service.validate = AsyncMock()
        service.get_schema = AsyncMock()
        depends.set(ValidationService, service)
        return service

    def test_validate_input_decorator(self, mock_validation_service: Mock) -> None:
        """Test validate_input decorator."""
        # Mock successful validation
        mock_validation_service.validate.return_value = ValidationResult(
            field_name="name",
            value="validated_name",
            original_value="test_name",
            is_valid=True
        )

        schema = StringValidationSchema("name", min_length=1)

        @validate_input({"name": schema})
        async def test_function(name: str) -> str:
            return f"Hello, {name}!"

        # This test verifies the decorator is applied correctly
        assert hasattr(test_function, '__wrapped__')

    def test_validate_output_decorator(self, mock_validation_service: Mock) -> None:
        """Test validate_output decorator."""
        # Mock successful validation
        mock_validation_service.validate.return_value = ValidationResult(
            field_name="output",
            value="validated_output",
            original_value="test_output",
            is_valid=True
        )

        schema = StringValidationSchema("output", min_length=1)

        @validate_output(schema)
        async def test_function() -> str:
            return "test_output"

        # This test verifies the decorator is applied correctly
        assert hasattr(test_function, '__wrapped__')

    def test_sanitize_input_decorator(self, mock_validation_service: Mock) -> None:
        """Test sanitize_input decorator."""
        # Mock successful validation/sanitization
        mock_validation_service.validate.return_value = ValidationResult(
            field_name="content",
            value="sanitized_content",
            original_value="<script>test</script>",
            is_valid=True
        )

        @sanitize_input(fields=["content"])
        async def test_function(content: str) -> str:
            return content

        # This test verifies the decorator is applied correctly
        assert hasattr(test_function, '__wrapped__')

    def test_validate_schema_decorator(self, mock_validation_service: Mock) -> None:
        """Test validate_schema decorator."""
        # Mock schema retrieval and validation
        mock_schema = Mock()
        mock_validation_service.get_schema.return_value = mock_schema
        mock_validation_service.validate.return_value = ValidationResult(
            field_name="data",
            value={"validated": "data"},
            original_value={"test": "data"},
            is_valid=True
        )

        @validate_schema("test_schema")
        async def test_function(**data) -> dict:
            return data

        # This test verifies the decorator is applied correctly
        assert hasattr(test_function, '__wrapped__')

    def test_validate_contracts_decorator(self, mock_validation_service: Mock) -> None:
        """Test validate_contracts decorator."""
        input_contract = {"name": str, "age": int}
        output_contract = {"result": str}

        @validate_contracts(
            input_contract=input_contract,
            output_contract=output_contract
        )
        async def test_function(name: str, age: int) -> dict:
            return {"result": f"{name} is {age} years old"}

        # This test verifies the decorator is applied correctly
        assert hasattr(test_function, '__wrapped__')


class TestDecoratorIntegration:
    """Integration tests for decorators with actual validation logic."""

    @pytest.fixture
    async def validation_service(self) -> ValidationService:
        """Create real ValidationService for integration tests."""
        from acb.services.validation import ValidationSettings
        from acb.config import Config
        from acb.logger import Logger

        # Mock dependencies
        config = Mock(spec=Config)
        logger = Mock(spec=Logger)
        logger.info = Mock()
        logger.warning = Mock()
        logger.error = Mock()

        depends.set(Config, config)
        depends.set(Logger, logger)

        settings = ValidationSettings(models_adapter_enabled=False)
        service = ValidationService(validation_settings=settings)
        await service.initialize()

        depends.set(ValidationService, service)
        return service

    async def test_validate_input_integration(self, validation_service: ValidationService) -> None:
        """Test validate_input decorator with real validation."""
        schema = StringValidationSchema("name", min_length=2)

        @validate_input({"name": schema})
        async def greet_user(name: str) -> str:
            return f"Hello, {name}!"

        # Test with valid input
        result = await greet_user("John")
        assert result == "Hello, John!"

        # Test with invalid input (should raise ValidationError)
        with pytest.raises(ValidationError):
            await greet_user("X")  # Too short

    async def test_validate_output_integration(self, validation_service: ValidationService) -> None:
        """Test validate_output decorator with real validation."""
        schema = StringValidationSchema("output", min_length=5)

        @validate_output(schema)
        async def get_message() -> str:
            return "Hello"

        # Test with valid output
        result = await get_message()
        assert result == "Hello"

        # Test with invalid output
        @validate_output(schema)
        async def get_short_message() -> str:
            return "Hi"  # Too short

        with pytest.raises(ValidationError):
            await get_short_message()

    async def test_sanitize_input_integration(self, validation_service: ValidationService) -> None:
        """Test sanitize_input decorator with real sanitization."""
        @sanitize_input(fields=["content"])
        async def process_content(content: str) -> str:
            return f"Processed: {content}"

        # Test with content that needs sanitization
        result = await process_content("<script>alert('xss')</script>Hello")
        assert "script" not in result.lower()
        assert "Hello" in result

    async def test_validate_schema_integration(self, validation_service: ValidationService) -> None:
        """Test validate_schema decorator with real schema."""
        # Register a schema
        schema = StringValidationSchema("test_data", min_length=3)
        validation_service.register_schema(schema)

        @validate_schema("test_data")
        async def process_data(data: str) -> str:
            return f"Processed: {data}"

        # Test with valid data
        result = await process_data("valid")
        assert result == "Processed: valid"

        # Test with invalid data
        with pytest.raises(ValidationError):
            await process_data("no")  # Too short

        # Test with non-existent schema
        @validate_schema("nonexistent_schema")
        async def process_with_missing_schema(data: str) -> str:
            return data

        with pytest.raises(ValidationError):
            await process_with_missing_schema("test")

    async def test_validate_contracts_integration(self, validation_service: ValidationService) -> None:
        """Test validate_contracts decorator with real contract validation."""
        @validate_contracts(
            input_contract={"name": str, "age": int},
            output_contract={"message": str, "age": int}
        )
        async def create_profile(name: str, age: int) -> dict:
            return {"message": f"User {name}", "age": age}

        # Test with valid input and output
        result = await create_profile("John", 30)
        assert result["message"] == "User John"
        assert result["age"] == 30

        # Test with invalid input type
        with pytest.raises(ValidationError):
            await create_profile("John", "thirty")  # age should be int

        # Test with invalid output contract
        @validate_contracts(
            input_contract={"name": str},
            output_contract={"result": int}  # Expects int but returns str
        )
        async def get_name_length(name: str) -> dict:
            return {"result": name}  # Returns string instead of int

        with pytest.raises(ValidationError):
            await get_name_length("test")

    async def test_decorator_chaining(self, validation_service: ValidationService) -> None:
        """Test chaining multiple validation decorators."""
        input_schema = StringValidationSchema("name", min_length=2)
        output_schema = StringValidationSchema("output", min_length=5)

        @validate_input({"name": input_schema})
        @validate_output(output_schema)
        @sanitize_input(fields=["name"])
        async def process_name(name: str) -> str:
            return f"Hello, {name}!"

        # Test successful processing through all decorators
        result = await process_name("John")
        assert result == "Hello, John!"

        # Test failure at input validation
        with pytest.raises(ValidationError):
            await process_name("X")  # Too short for input schema

    async def test_sync_function_decoration(self, validation_service: ValidationService) -> None:
        """Test decorator application to synchronous functions."""
        schema = StringValidationSchema("text", min_length=3)

        @validate_input({"text": schema})
        def sync_function(text: str) -> str:
            return f"Processed: {text}"

        # Test that sync function is properly wrapped
        assert hasattr(sync_function, '__wrapped__')

        # Note: Actually calling the sync function would require running asyncio.run
        # internally, which is complex to test. The important part is that the
        # decorator is applied correctly.



"""Queue provider discovery system for ACB framework.

This module provides dynamic queue provider discovery and loading capabilities,
following ACB's adapter patterns with capability-based selection and
configuration overrides.
"""

import logging
import typing as t
from contextlib import suppress
from enum import Enum
from pathlib import Path
from uuid import UUID, uuid4

from pydantic import BaseModel, Field
from acb.config import Config
from acb.depends import depends

from ._base import QueueBase, QueueCapability, QueueMetadata, QueueSettings

logger = logging.getLogger(__name__)


class QueueProviderStatus(Enum):
    """Queue provider availability status."""

    AVAILABLE = "available"
    MISSING_DEPENDENCIES = "missing_dependencies"
    NOT_INSTALLED = "not_installed"
    DISABLED = "disabled"
    ERROR = "error"


class QueueProviderDescriptor(BaseModel):
    """Descriptor for a queue provider."""

    provider_id: UUID = Field(description="Unique provider identifier")
    name: str = Field(description="Provider name")
    module_path: str = Field(description="Python module path")
    class_name: str = Field(description="Queue class name")
    factory_function: str | None = Field(
        default=None,
        description="Factory function name",
    )

    # Metadata
    metadata: QueueMetadata | None = Field(
        default=None,
        description="Provider metadata",
    )
    status: QueueProviderStatus = Field(default=QueueProviderStatus.AVAILABLE)
    error_message: str | None = Field(
        default=None,
        description="Error message if unavailable",
    )

    # Configuration
    default_settings: dict[str, t.Any] = Field(default_factory=dict)
    config_schema: dict[str, t.Any] = Field(default_factory=dict)

    # Dependencies
    required_packages: list[str] = Field(default_factory=list)
    optional_packages: list[str] = Field(default_factory=list)


class QueueProviderNotFound(Exception):
    """Exception raised when a queue provider is not found."""

    def __init__(self, provider_name: str) -> None:
        self.provider_name = provider_name
        super().__init__(f"Queue provider not found: {provider_name}")


class QueueProviderNotInstalled(Exception):
    """Exception raised when a queue provider is not installed."""

    def __init__(self, provider_name: str, missing_packages: list[str]) -> None:
        self.provider_name = provider_name
        self.missing_packages = missing_packages
        super().__init__(
            f"Queue provider {provider_name} is not installed. "
            f"Missing packages: {', '.join(missing_packages)}",
        )


# Global provider registry
_provider_registry: dict[str, QueueProviderDescriptor] = {}
_provider_overrides: dict[str, str] = {}
_initialized = False


def generate_provider_id() -> UUID:
    """Generate a unique provider identifier.

    Returns:
        UUID7 identifier
    """
    return uuid4()


def create_queue_metadata_template() -> QueueMetadata:
    """Create a template QueueMetadata instance.

    Returns:
        QueueMetadata template
    """
    return QueueMetadata(
        queue_id=generate_provider_id(),
        name="Custom Queue Provider",
        description="Custom queue provider template",
        version="1.0.0",
        capabilities=[QueueCapability.BASIC_QUEUE],
    )


def register_queue_providers() -> None:
    """Register built-in queue providers."""
    global _provider_registry, _initialized

    if _initialized:
        return

    # Memory queue provider
    _provider_registry["memory"] = QueueProviderDescriptor(
        provider_id=generate_provider_id(),
        name="Memory Queue",
        module_path="acb.queues.memory",
        class_name="MemoryQueue",
        factory_function="create_memory_queue",
        required_packages=[],
        default_settings={
            "max_memory_usage": 100_000_000,
            "max_tasks_per_queue": 10_000,
        },
    )

    # Redis queue provider
    _provider_registry["redis"] = QueueProviderDescriptor(
        provider_id=generate_provider_id(),
        name="Redis Queue",
        module_path="acb.queues.redis",
        class_name="RedisQueue",
        factory_function="create_redis_queue",
        required_packages=["redis>=5.0.0"],
        default_settings={
            "redis_url": "redis://localhost:6379/0",
            "key_prefix": "acb:queue",
        },
    )

    # RabbitMQ queue provider
    _provider_registry["rabbitmq"] = QueueProviderDescriptor(
        provider_id=generate_provider_id(),
        name="RabbitMQ Queue",
        module_path="acb.queues.rabbitmq",
        class_name="RabbitMQQueue",
        factory_function="create_rabbitmq_queue",
        required_packages=["aio-pika>=9.0.0"],
        default_settings={
            "rabbitmq_url": "amqp://localhost:5672/",
            "exchange_name": "acb.tasks",
        },
    )

    _initialized = True
    logger.debug("Queue providers registered")


def _check_provider_dependencies(
    descriptor: QueueProviderDescriptor,
) -> tuple[QueueProviderStatus, str | None]:
    """Check if provider dependencies are available.

    Args:
        descriptor: Provider descriptor

    Returns:
        Tuple of (status, error_message)
    """
    import importlib.util

    missing_packages = []

    for package in descriptor.required_packages:
        # Extract package name (handle version specs)
        package_name = package.split(">=")[0].split("==")[0].split("<")[0].split(">")[0]

        spec = importlib.util.find_spec(package_name)
        if spec is None:
            missing_packages.append(package)

    if missing_packages:
        return (
            QueueProviderStatus.MISSING_DEPENDENCIES,
            f"Missing packages: {', '.join(missing_packages)}",
        )

    return QueueProviderStatus.AVAILABLE, None


def _load_provider_metadata(
    descriptor: QueueProviderDescriptor,
) -> QueueMetadata | None:
    """Load provider metadata from module.

    Args:
        descriptor: Provider descriptor

    Returns:
        QueueMetadata if available
    """
    try:
        import importlib

        module = importlib.import_module(descriptor.module_path)
        if hasattr(module, "MODULE_METADATA"):
            return module.MODULE_METADATA
    except Exception as e:
        logger.debug(f"Failed to load metadata for {descriptor.name}: {e}")

    return None


def get_queue_provider_class(provider_name: str) -> type[QueueBase]:
    """Get queue provider class by name.

    Args:
        provider_name: Provider name

    Returns:
        Queue class

    Raises:
        QueueProviderNotFound: If provider not found
        QueueProviderNotInstalled: If provider dependencies missing
    """
    register_queue_providers()

    # Check for override
    actual_provider = _provider_overrides.get(provider_name, provider_name)

    if actual_provider not in _provider_registry:
        raise QueueProviderNotFound(actual_provider)

    descriptor = _provider_registry[actual_provider]

    # Check dependencies
    status, _error_msg = _check_provider_dependencies(descriptor)
    if status == QueueProviderStatus.MISSING_DEPENDENCIES:
        raise QueueProviderNotInstalled(actual_provider, descriptor.required_packages)

    try:
        import importlib

        module = importlib.import_module(descriptor.module_path)
        queue_class = getattr(module, descriptor.class_name)

        # Update metadata if available
        if descriptor.metadata is None:
            descriptor.metadata = _load_provider_metadata(descriptor)

        return queue_class

    except ImportError as e:
        raise QueueProviderNotInstalled(
            actual_provider,
            descriptor.required_packages,
        ) from e
    except AttributeError as e:
        raise QueueProviderNotFound(actual_provider) from e


def get_queue_provider_descriptor(provider_name: str) -> QueueProviderDescriptor:
    """Get queue provider descriptor by name.

    Args:
        provider_name: Provider name

    Returns:
        QueueProviderDescriptor

    Raises:
        QueueProviderNotFound: If provider not found
    """
    register_queue_providers()

    # Check for override
    actual_provider = _provider_overrides.get(provider_name, provider_name)

    if actual_provider not in _provider_registry:
        raise QueueProviderNotFound(actual_provider)

    descriptor = _provider_registry[actual_provider]

    # Update status and metadata
    status, error_msg = _check_provider_dependencies(descriptor)
    descriptor.status = status
    descriptor.error_message = error_msg

    if descriptor.metadata is None:
        descriptor.metadata = _load_provider_metadata(descriptor)

    return descriptor


def get_queue_provider_info(provider_name: str) -> dict[str, t.Any]:
    """Get comprehensive information about a queue provider.

    Args:
        provider_name: Provider name

    Returns:
        Provider information dictionary
    """
    try:
        descriptor = get_queue_provider_descriptor(provider_name)

        info = {
            "name": descriptor.name,
            "provider_id": str(descriptor.provider_id),
            "module_path": descriptor.module_path,
            "class_name": descriptor.class_name,
            "status": descriptor.status.value,
            "required_packages": descriptor.required_packages,
            "optional_packages": descriptor.optional_packages,
            "default_settings": descriptor.default_settings,
        }

        if descriptor.error_message:
            info["error_message"] = descriptor.error_message

        if descriptor.metadata:
            info["metadata"] = {
                "version": descriptor.metadata.version,
                "description": descriptor.metadata.description,
                "capabilities": [cap.value for cap in descriptor.metadata.capabilities],
                "max_throughput": descriptor.metadata.max_throughput,
                "max_workers": descriptor.metadata.max_workers,
                "supports_clustering": descriptor.metadata.supports_clustering,
            }

        return info

    except QueueProviderNotFound:
        return {
            "name": provider_name,
            "status": QueueProviderStatus.NOT_INSTALLED.value,
            "error_message": f"Provider {provider_name} not found",
        }


def import_queue_provider(provider_name: str | None = None) -> type[QueueBase]:
    """Import a queue provider class with error handling.

    Args:
        provider_name: Provider name (default: from config)

    Returns:
        Queue class

    Raises:
        QueueProviderNotFound: If provider not found
        QueueProviderNotInstalled: If provider dependencies missing
    """
    if provider_name is None:
        # Get from configuration
        config = depends.get(Config)
        provider_name = getattr(config, "queue_provider", "memory")

    return get_queue_provider_class(provider_name or "memory")


def try_import_queue_provider(provider_name: str) -> type[QueueBase] | None:
    """Try to import a queue provider class without raising exceptions.

    Args:
        provider_name: Provider name

    Returns:
        Queue class or None if unavailable
    """
    try:
        return import_queue_provider(provider_name)
    except (QueueProviderNotFound, QueueProviderNotInstalled):
        return None


def list_queue_providers() -> list[str]:
    """List all available queue provider names.

    Returns:
        List of provider names
    """
    register_queue_providers()
    return list(_provider_registry.keys())


def list_available_queue_providers() -> list[str]:
    """List queue providers that have all dependencies available.

    Returns:
        List of available provider names
    """
    register_queue_providers()
    available = []

    for name, descriptor in _provider_registry.items():
        status, _ = _check_provider_dependencies(descriptor)
        if status == QueueProviderStatus.AVAILABLE:
            available.append(name)

    return available


def list_enabled_queue_providers() -> list[str]:
    """List enabled queue providers (same as available for now).

    Returns:
        List of enabled provider names
    """
    return list_available_queue_providers()


def list_queue_providers_by_capability(capability: QueueCapability) -> list[str]:
    """List queue providers that support a specific capability.

    Args:
        capability: Required capability

    Returns:
        List of provider names
    """
    register_queue_providers()
    matching = []

    for name, descriptor in _provider_registry.items():
        if descriptor.metadata and capability in descriptor.metadata.capabilities:
            # Check if dependencies are available
            status, _ = _check_provider_dependencies(descriptor)
            if status == QueueProviderStatus.AVAILABLE:
                matching.append(name)

    return matching


def enable_queue_provider(provider_name: str, target_provider: str) -> None:
    """Enable a queue provider override.

    Args:
        provider_name: Logical provider name
        target_provider: Actual provider to use
    """
    global _provider_overrides
    _provider_overrides[provider_name] = target_provider
    logger.info(
        f"Enabled queue provider override: {provider_name} -> {target_provider}",
    )


def disable_queue_provider(provider_name: str) -> None:
    """Disable a queue provider override.

    Args:
        provider_name: Provider name to disable override for
    """
    global _provider_overrides
    if provider_name in _provider_overrides:
        target = _provider_overrides.pop(provider_name)
        logger.info(f"Disabled queue provider override: {provider_name} -> {target}")


def get_queue_provider_override(provider_name: str) -> str | None:
    """Get the override target for a provider.

    Args:
        provider_name: Provider name

    Returns:
        Override target or None
    """
    return _provider_overrides.get(provider_name)


def apply_queue_provider_overrides(config_path: str | Path | None = None) -> None:
    """Apply queue provider overrides from configuration.

    Args:
        config_path: Path to configuration file
    """
    if config_path is None:
        config_path = Path("settings/queues.yaml")
    else:
        config_path = Path(config_path)

    if not config_path.exists():
        logger.debug(f"Queue configuration file not found: {config_path}")
        return

    try:
        import yaml

        with config_path.open() as f:
            config = yaml.safe_load(f)

        overrides = config.get("provider_overrides", {})
        for provider_name, target_provider in overrides.items():
            enable_queue_provider(provider_name, target_provider)

        logger.info(f"Applied queue provider overrides from {config_path}")

    except Exception as e:
        logger.exception(f"Failed to apply queue provider overrides: {e}")


# Factory functions
def create_queue_instance(
    provider_name: str | None = None,
    settings: QueueSettings | None = None,
    **kwargs: t.Any,
) -> QueueBase:
    """Create a queue instance.

    Args:
        provider_name: Provider name (default: from config)
        settings: Queue settings
        **kwargs: Additional settings

    Returns:
        Queue instance
    """
    queue_class = import_queue_provider(provider_name)

    if settings is None:
        # Try to get provider-specific settings class
        with suppress(Exception):
            descriptor = get_queue_provider_descriptor(provider_name or "memory")
            if descriptor.default_settings:
                import importlib

                module = importlib.import_module(descriptor.module_path)
                settings_class_name = f"{descriptor.class_name}Settings"

                if hasattr(module, settings_class_name):
                    settings_class = getattr(module, settings_class_name)
                    settings = settings_class(**descriptor.default_settings, **kwargs)

    return queue_class(settings)


async def create_queue_instance_async(
    provider_name: str | None = None,
    settings: QueueSettings | None = None,
    **kwargs: t.Any,
) -> QueueBase:
    """Create and start a queue instance asynchronously.

    Args:
        provider_name: Provider name (default: from config)
        settings: Queue settings
        **kwargs: Additional settings

    Returns:
        Started queue instance
    """
    queue = create_queue_instance(provider_name, settings, **kwargs)
    await queue.start()
    return queue


# Context manager for queue instances
class QueueContext:
    """Context manager for queue instances."""

    def __init__(
        self,
        provider_name: str | None = None,
        settings: QueueSettings | None = None,
        **kwargs: t.Any,
    ) -> None:
        self.provider_name = provider_name
        self.settings = settings
        self.kwargs = kwargs
        self.queue: QueueBase | None = None

    async def __aenter__(self) -> QueueBase:
        """Enter the context manager."""
        self.queue = await create_queue_instance_async(
            self.provider_name,
            self.settings,
            **self.kwargs,
        )
        return self.queue

    async def __aexit__(self, exc_type: t.Any, exc_val: t.Any, exc_tb: t.Any) -> None:
        """Exit the context manager."""
        if self.queue:
            await self.queue.stop()


def queue_context(
    provider_name: str | None = None,
    settings: QueueSettings | None = None,
    **kwargs: t.Any,
) -> QueueContext:
    """Create a queue context manager.

    Args:
        provider_name: Provider name (default: from config)
        settings: Queue settings
        **kwargs: Additional settings

    Returns:
        QueueContext instance
    """
    return QueueContext(provider_name, settings, **kwargs)


# Auto-discovery initialization
def initialize_queue_discovery() -> None:
    """Initialize queue provider discovery."""
    register_queue_providers()

    # Apply configuration overrides
    try:
        apply_queue_provider_overrides()
    except Exception as e:
        logger.debug(f"Failed to apply queue provider overrides: {e}")

    logger.debug("Queue provider discovery initialized")


# Initialize on import
initialize_queue_discovery()



"""Redis Queue Backend Adapter for ACB.

High-performance Redis-backed queue implementation suitable for production
deployments with persistence, clustering, and horizontal scaling support.

Features:
    - Connection pooling with configurable limits
    - Priority queue support using Redis sorted sets
    - Delayed message delivery with timestamp-based scheduling
    - Dead letter queue for failed message handling
    - Batch operations for improved throughput
    - Pub/sub support for event-driven messaging
    - Atomic operations using Lua scripts
    - Health monitoring and automatic reconnection

Requirements:
    - Redis server (standalone or cluster)
    - coredis for async Redis client

Example:
    Basic task queue usage:

    ```python
    from acb.depends import Inject, depends
    from acb.adapters import import_adapter

    Queue = import_adapter("queue")


    @depends.inject
    async def process_tasks(queue: Queue = depends()):
        # Enqueue task with priority
        await queue.enqueue(
            "tasks", b"task payload", priority=MessagePriority.HIGH, delay_seconds=10
        )

        # Dequeue and process
        message = await queue.dequeue("tasks")
        if message:
            # Process task
            await queue.acknowledge(message)
    ```

    Pub/sub pattern:

    ```python
    # Publisher
    await queue.publish("events.user", b"user.created")

    # Subscriber
    async with queue.subscribe("events.*") as messages:
        async for message in messages:
            await process_event(message)
            await queue.acknowledge(message)
    ```

Author: Claude Code
Created: 2025-10-01
"""

import asyncio
import time
import typing as t
from collections.abc import AsyncGenerator, AsyncIterator
from contextlib import asynccontextmanager, suppress
from uuid import UUID

from pydantic import Field
from acb.adapters import AdapterCapability, AdapterMetadata, AdapterStatus
from acb.cleanup import CleanupMixin
from acb.config import Config
from acb.depends import Inject, depends

LoggerType = t.Any

from ._base import (
    MessagePriority,
    MessagingCapability,
    MessagingConnectionError,
    MessagingOperationError,
    MessagingTimeoutError,
    PubSubMessage,
    QueueMessage,
    Subscription,
)
from ._base import (
    MessagingSettings as BaseMessagingSettings,
)

# Lazy imports for coredis
_coredis_imports: dict[str, t.Any] = {}

MODULE_METADATA = AdapterMetadata(
    module_id=UUID("fef2289a-83f8-43e3-a5c0-a5343ee8bc88"),
    name="Redis Messaging",
    category="messaging",
    provider="redis",
    version="1.0.0",
    acb_min_version="0.19.0",
    author="Claude Code",
    created_date="2025-10-01",
    last_modified="2025-10-08",
    status=AdapterStatus.STABLE,
    capabilities=[
        AdapterCapability.ASYNC_OPERATIONS,
        AdapterCapability.CONNECTION_POOLING,
        AdapterCapability.CACHING,
        AdapterCapability.TRANSACTIONS,
        AdapterCapability.HEALTH_CHECKS,
        AdapterCapability.RECONNECTION,
    ],
    required_packages=["coredis>=4.0.0"],
    description="High-performance Redis messaging backend with pub/sub and queue support",
    settings_class="RedisMessagingSettings",
    config_example={
        "connection_url": "redis://localhost:6379/0",
        "max_connections": 20,
        "key_prefix": "acb:messaging",
        "enable_clustering": False,
        "use_lua_scripts": True,
    },
)


def _get_coredis_imports() -> dict[str, t.Any]:
    """Lazy import of coredis dependencies."""
    if not _coredis_imports:
        try:
            from coredis import Redis, RedisCluster
            from coredis.exceptions import (
                ConnectionError,
                RedisError,
                TimeoutError,
            )
            from coredis.pool import ConnectionPool

            _coredis_imports.update(
                {
                    "Redis": Redis,
                    "RedisCluster": RedisCluster,
                    "ConnectionPool": ConnectionPool,
                    "ConnectionError": ConnectionError,
                    "RedisError": RedisError,
                    "TimeoutError": TimeoutError,
                }
            )
        except ImportError as e:
            raise ImportError(
                "coredis is required for RedisQueue. "
                "Install with: pip install coredis>=4.0.0"
            ) from e

    return _coredis_imports


class RedisMessagingSettings(BaseMessagingSettings):
    """Settings for Redis messaging implementation."""

    # Redis connection
    connection_url: str | None = "redis://localhost:6379/0"
    key_prefix: str = "acb:messaging"

    # Connection pool settings
    max_connections: int = 20
    socket_connect_timeout: float = 5.0
    socket_timeout: float = 5.0

    # Clustering
    enable_clustering: bool = False
    cluster_nodes: list[str] = Field(default_factory=list)

    # Performance
    use_lua_scripts: bool = True
    pipeline_size: int = 100

    # Message retention
    message_ttl: int = 86400  # 24 hours

    # Operation timeouts
    send_timeout: float = 10.0
    ack_timeout: float = 5.0
    dead_letter_ttl: int = 604800  # 7 days

    @depends.inject
    def __init__(self, config: Inject[Config], **values: t.Any) -> None:
        super().__init__(**values)


class RedisMessaging(CleanupMixin):
    """Redis-backed unified messaging implementation.

    Provides high-performance messaging operations using Redis data structures:
    - Pub/sub for event-driven messaging patterns
    - Sorted sets for priority queues and delayed messages
    - Lists for FIFO queue operations
    - Hash maps for message metadata storage
    """

    config: Config

    def __init__(self, settings: RedisMessagingSettings | None = None) -> None:
        """Initialize Redis messaging backend.

        Args:
            settings: Redis messaging configuration
        """
        super().__init__()
        self._settings: RedisMessagingSettings = settings or RedisMessagingSettings()

        # Redis client and pool
        self._client: t.Any = None
        self._connection_pool: t.Any = None

        # Connection management attributes (required by base class)
        self._connection_lock = asyncio.Lock()
        self._shutdown_event = asyncio.Event()

        # Key patterns for Redis operations
        self._queue_key = f"{self._settings.key_prefix}:queues:{{topic}}"
        self._delayed_key = f"{self._settings.key_prefix}:delayed"
        self._processing_key = f"{self._settings.key_prefix}:processing"
        self._dead_letter_key = f"{self._settings.key_prefix}:dead_letter"
        self._message_key = f"{self._settings.key_prefix}:messages:{{message_id}}"
        self._pubsub_key = f"{self._settings.key_prefix}:pubsub:{{topic}}"

        # Lua scripts for atomic operations
        self._lua_scripts: dict[str, t.Any] = {}

        # Background task handles
        self._delayed_processor_task: asyncio.Task[None] | None = None
        self._pubsub_client: t.Any = None
        self._logger: LoggerType | None = None
        self._connected: bool = False

    @property
    def logger(self) -> LoggerType:
        """Lazy-initialize logger.

        Returns:
            Logger instance
        """
        if self._logger is None:
            from acb.adapters import import_adapter

            logger_cls = import_adapter("logger")
            self._logger = t.cast(LoggerType, depends.get_sync(logger_cls))
        return self._logger

    # ========================================================================
    # Connection Management (Private Implementation)
    # ========================================================================

    async def _ensure_client(self) -> t.Any:
        """Ensure Redis client is initialized (lazy initialization).

        Returns:
            Redis client instance

        Raises:
            MessagingConnectionError: If connection fails
        """
        if self._client is None:
            async with self._connection_lock:
                # Double-check after acquiring lock
                if self._client is None:
                    imports = _get_coredis_imports()

                    try:
                        # Create connection pool
                        self._connection_pool = imports["ConnectionPool"].from_url(
                            self._settings.connection_url,
                            max_connections=self._settings.max_connections,
                            socket_connect_timeout=self._settings.socket_connect_timeout,
                            socket_timeout=self._settings.socket_timeout,
                            decode_responses=False,  # We handle encoding
                        )

                        # Create Redis client
                        if self._settings.enable_clustering:
                            self._client = imports["RedisCluster"](
                                connection_pool=self._connection_pool
                            )
                        else:
                            self._client = imports["Redis"](
                                connection_pool=self._connection_pool
                            )

                        # Register for cleanup
                        self.register_resource(self._client)

                        # Test connection
                        await self._client.ping()  # type: ignore[attr-defined]

                        self.logger.debug("Redis connection established")

                    except Exception as e:
                        self.logger.exception(f"Failed to connect to Redis: {e}")
                        raise MessagingConnectionError(
                            "Failed to establish Redis connection",
                            original_error=e,
                        ) from e

        return self._client

    async def _connect(self) -> None:
        """Establish connection to Redis backend."""
        async with self._connection_lock:
            if self._connected:
                return

            # Initialize client
            await self._ensure_client()

            # Load Lua scripts if enabled
            if self._settings.use_lua_scripts:
                await self._load_lua_scripts()

            # Start background tasks
            self._delayed_processor_task = asyncio.create_task(
                self._process_delayed_messages()
            )

            self._connected = True
            self.logger.info("Redis queue backend connected")

    async def _disconnect(self) -> None:
        """Disconnect from Redis backend."""
        self._connected = False
        self._shutdown_event.set()

        # Cancel background tasks
        if self._delayed_processor_task:
            self._delayed_processor_task.cancel()
            with suppress(asyncio.CancelledError):
                await self._delayed_processor_task

        # Close pub/sub client if active
        if self._pubsub_client:
            with suppress(Exception):
                await self._pubsub_client.close()

        # Close main client
        if self._client:
            with suppress(Exception):
                await self._client.close()

            self._client = None

        # Close connection pool
        if self._connection_pool:
            with suppress(Exception):
                await self._connection_pool.disconnect()

            self._connection_pool = None

        # Cleanup resources
        await self.cleanup()

        self.logger.info("Redis queue backend disconnected")

    async def _health_check(self) -> dict[str, t.Any]:
        """Perform Redis health check.

        Returns:
            Health status information
        """
        try:
            client = await self._ensure_client()

            # Measure latency
            start_time = time.time()
            await client.ping()
            latency_ms = (time.time() - start_time) * 1000

            # Get Redis info
            info = await client.info()

            return {
                "healthy": True,
                "connected": self._connected,
                "latency_ms": latency_ms,
                "backend_info": {
                    "redis_version": info.get(b"redis_version", b"").decode(),
                    "used_memory": info.get(b"used_memory", 0),
                    "connected_clients": info.get(b"connected_clients", 0),
                    "uptime_seconds": info.get(b"uptime_in_seconds", 0),
                },
            }

        except Exception as e:
            self.logger.exception(f"Redis health check failed: {e}")
            return {
                "healthy": False,
                "connected": False,
                "error": str(e),
            }

    # ========================================================================
    # Lua Scripts
    # ========================================================================

    async def _load_lua_scripts(self) -> None:
        """Load Lua scripts for atomic operations."""
        client = await self._ensure_client()

        try:
            # Script for atomic enqueue with priority
            enqueue_script = """
            local queue_key = KEYS[1]
            local message_key = KEYS[2]
            local message_data = ARGV[1]
            local score = tonumber(ARGV[2])

            -- Store message data
            redis.call('SET', message_key, message_data)
            redis.call('EXPIRE', message_key, tonumber(ARGV[3]))

            -- Add to sorted set with score (timestamp + priority)
            redis.call('ZADD', queue_key, score, message_key)

            return 1
            """

            # Script for atomic dequeue
            dequeue_script = """
            local queue_key = KEYS[1]
            local processing_key = KEYS[2]
            local current_time = tonumber(ARGV[1])

            -- Get messages ready for processing
            local messages = redis.call('ZRANGEBYSCORE', queue_key, '-inf', current_time, 'LIMIT', 0, 1)

            if #messages == 0 then
                return nil
            end

            local message_key = messages[1]

            -- Move to processing atomically
            redis.call('ZREM', queue_key, message_key)
            redis.call('ZADD', processing_key, current_time, message_key)

            -- Get message data
            local message_data = redis.call('GET', message_key)
            return {message_key, message_data}
            """

            # Script for message acknowledgment
            ack_script = """
            local processing_key = KEYS[1]
            local message_key = ARGV[1]

            -- Remove from processing
            redis.call('ZREM', processing_key, message_key)

            -- Delete message data (or keep based on retention policy)
            redis.call('DEL', message_key)

            return 1
            """

            # Register scripts
            self._lua_scripts = {
                "enqueue": await client.script_load(enqueue_script.encode()),
                "dequeue": await client.script_load(dequeue_script.encode()),
                "ack": await client.script_load(ack_script.encode()),
            }

            self.logger.debug("Lua scripts loaded successfully")

        except Exception as e:
            self.logger.warning(f"Failed to load Lua scripts: {e}")
            self._settings.use_lua_scripts = False

    # ========================================================================
    # Message Operations (Private Implementation)
    # ========================================================================

    async def _send(
        self,
        message: QueueMessage,
        timeout: float | None = None,
    ) -> str:
        """Send a message (private implementation).

        Args:
            message: Message to send
            timeout: Optional timeout override

        Returns:
            Message ID

        Raises:
            MessagingConnectionError: If not connected
            MessagingOperationError: If send fails
            QueueTimeoutError: If operation times out
        """
        if not self._connected:
            raise MessagingConnectionError("Not connected to Redis")

        client = await self._ensure_client()
        timeout = timeout or self._settings.send_timeout

        try:
            # Serialize message using Pydantic
            message_data = message.model_dump_json().encode()
            message_key = self._message_key.format(message_id=message.message_id)

            # Calculate score for sorted set
            # Score = scheduled_time * 1M - priority (higher priority = lower score)
            scheduled_time = time.time()
            if message.delay_seconds > 0:
                scheduled_time += message.delay_seconds

            score = scheduled_time * 1_000_000 - message.priority.value

            # Determine target queue
            if message.delay_seconds > 0:
                queue_key = self._delayed_key
            else:
                queue_key = self._queue_key.format(topic=message.queue)

            # Use Lua script for atomic operation if available
            if self._settings.use_lua_scripts and "enqueue" in self._lua_scripts:
                await asyncio.wait_for(
                    client.evalsha(
                        self._lua_scripts["enqueue"],
                        3,
                        queue_key.encode(),
                        message_key.encode(),
                        message_data,
                        str(score).encode(),
                        str(self._settings.message_ttl).encode(),
                    ),
                    timeout=timeout,
                )
            else:
                # Use pipeline for atomic operation
                async with client.pipeline() as pipe:
                    await pipe.set(
                        message_key.encode(),
                        message_data,
                        ex=self._settings.message_ttl,
                    )
                    await pipe.zadd(queue_key.encode(), {message_key.encode(): score})
                    await asyncio.wait_for(pipe.execute(), timeout=timeout)

            self.logger.debug(
                f"Sent message {message.message_id} to queue {message.queue}"
            )
            return str(message.message_id)

        except TimeoutError as e:
            raise MessagingTimeoutError(
                f"Send operation timed out after {timeout}s",
                original_error=e,
            ) from e
        except Exception as e:
            self.logger.exception(f"Failed to send message: {e}")
            raise MessagingOperationError(
                "Failed to send message",
                original_error=e,
            ) from e

    async def _receive_with_lua_script(
        self,
        client: t.Any,
        queue_key: str,
        current_time: float,
        timeout: float,
    ) -> QueueMessage | None:
        """Receive message using Lua script for atomic operation."""
        result = await asyncio.wait_for(
            client.evalsha(
                self._lua_scripts["dequeue"],
                2,
                queue_key.encode(),
                self._processing_key.encode(),
                str(current_time).encode(),
            ),
            timeout=timeout,
        )

        if result is None:
            return None

        message_key, message_data = result
        return QueueMessage.model_validate_json(message_data)

    async def _receive_with_manual_atomic(
        self,
        client: t.Any,
        queue_key: str,
        current_time: float,
        timeout: float,
    ) -> QueueMessage | None:
        """Receive message using manual atomic operation with pipeline."""
        # Get available messages
        messages = await client.zrangebyscore(
            queue_key.encode(),
            b"-inf",
            str(current_time).encode(),
            start=0,
            num=1,
        )

        if not messages:
            return None

        message_key = messages[0]

        # Move to processing queue atomically
        async with client.pipeline() as pipe:
            await pipe.zrem(queue_key.encode(), message_key)
            await pipe.zadd(
                self._processing_key.encode(),
                {message_key: time.time()},
            )
            await pipe.get(message_key)
            results = await asyncio.wait_for(pipe.execute(), timeout=timeout)

        message_data = results[-1]
        if message_data:
            return QueueMessage.model_validate_json(message_data)

        return None

    async def _receive(
        self,
        topic: str,
        timeout: float | None = None,
    ) -> QueueMessage | None:
        """Receive a message (private implementation).

        Args:
            topic: Topic/queue to receive from
            timeout: Optional timeout override

        Returns:
            Message or None if no messages available

        Raises:
            MessagingConnectionError: If not connected
            MessagingOperationError: If receive fails
        """
        if not self._connected:
            raise MessagingConnectionError("Not connected to Redis")

        client = await self._ensure_client()
        timeout = timeout or self._settings.connection_timeout
        queue_key = self._queue_key.format(topic=topic)

        try:
            current_time = time.time() * 1_000_000

            # Use Lua script for atomic operation if available
            if self._settings.use_lua_scripts and "dequeue" in self._lua_scripts:
                return await self._receive_with_lua_script(
                    client, queue_key, current_time, timeout
                )

            # Manual atomic operation using pipeline
            return await self._receive_with_manual_atomic(
                client, queue_key, current_time, timeout
            )

        except TimeoutError:
            return None  # Timeout is expected for blocking receives
        except Exception as e:
            self.logger.exception(f"Failed to receive message: {e}")
            raise MessagingOperationError(
                "Failed to receive message",
                original_error=e,
            ) from e

    async def _acknowledge(
        self,
        message: QueueMessage,
        timeout: float | None = None,
    ) -> None:
        """Acknowledge a message (private implementation).

        Args:
            message: Message to acknowledge
            timeout: Optional timeout override

        Raises:
            MessagingOperationError: If ack fails
        """
        if not self._connected:
            raise MessagingConnectionError("Not connected to Redis")

        client = await self._ensure_client()
        timeout = timeout or self._settings.connection_timeout
        message_key = self._message_key.format(message_id=message.message_id)

        try:
            # Use Lua script for atomic operation if available
            if self._settings.use_lua_scripts and "ack" in self._lua_scripts:
                await asyncio.wait_for(
                    client.evalsha(
                        self._lua_scripts["ack"],
                        1,
                        self._processing_key.encode(),
                        message_key.encode(),
                    ),
                    timeout=timeout,
                )
            else:
                # Manual operation
                async with client.pipeline() as pipe:
                    await pipe.zrem(self._processing_key.encode(), message_key.encode())
                    await pipe.delete(message_key.encode())
                    await asyncio.wait_for(pipe.execute(), timeout=timeout)

            self.logger.debug(f"Acknowledged message {message.message_id}")

        except Exception as e:
            self.logger.exception(f"Failed to acknowledge message: {e}")
            raise MessagingOperationError(
                "Failed to acknowledge message",
                original_error=e,
            ) from e

    async def _reject(
        self,
        message: QueueMessage,
        requeue: bool = False,
        timeout: float | None = None,
    ) -> None:
        """Reject a message (private implementation).

        Args:
            message: Message to reject
            requeue: Whether to requeue the message
            timeout: Optional timeout override

        Raises:
            MessagingOperationError: If reject fails
        """
        if not self._connected:
            raise MessagingConnectionError("Not connected to Redis")

        client = await self._ensure_client()
        timeout = timeout or self._settings.ack_timeout
        message_key = self._message_key.format(message_id=message.message_id)

        try:
            async with client.pipeline() as pipe:
                # Remove from processing
                await pipe.zrem(self._processing_key.encode(), message_key.encode())

                if requeue and message.retry_count < message.max_retries:
                    # Increment retry count and requeue
                    message.retry_count += 1
                    queue_key = self._queue_key.format(topic=message.queue)
                    score = time.time() * 1_000_000 - message.priority.value

                    await pipe.set(
                        message_key.encode(),
                        message.model_dump_json().encode(),
                        ex=self._settings.message_ttl,
                    )
                    await pipe.zadd(queue_key.encode(), {message_key.encode(): score})
                else:
                    # Move to dead letter queue
                    await pipe.zadd(
                        self._dead_letter_key.encode(),
                        {message_key.encode(): time.time()},
                    )
                    await pipe.expire(
                        message_key.encode(), self._settings.dead_letter_ttl
                    )

                await asyncio.wait_for(pipe.execute(), timeout=timeout)

            self.logger.debug(
                f"Rejected message {message.message_id} (requeue={requeue})"
            )

        except Exception as e:
            self.logger.exception(f"Failed to reject message: {e}")
            raise MessagingOperationError(
                "Failed to reject message",
                original_error=e,
            ) from e

    @asynccontextmanager  # type: ignore[arg-type]
    async def _subscribe(
        self,
        topic: str,
        prefetch: int | None = None,
    ) -> AsyncIterator[AsyncGenerator[QueueMessage]]:
        """Subscribe to topic (private implementation).

        Args:
            topic: Topic/pattern to subscribe to
            prefetch: Messages to prefetch

        Yields:
            Async generator of messages
        """
        if not self._connected:
            raise MessagingConnectionError("Not connected to Redis")

        client = await self._ensure_client()
        pubsub_channel = self._pubsub_key.format(topic=topic)

        # Create pub/sub client
        pubsub = client.pubsub()

        try:
            # Subscribe to channel/pattern
            if "*" in topic or "?" in topic:
                await pubsub.psubscribe(pubsub_channel.encode())
            else:
                await pubsub.subscribe(pubsub_channel.encode())

            async def message_generator() -> AsyncGenerator[QueueMessage]:
                """Generate messages from pub/sub."""
                with suppress(asyncio.CancelledError):
                    async for raw_message in pubsub.listen():
                        if raw_message[b"type"] in (b"message", b"pmessage"):
                            message_data = raw_message[b"data"]
                            try:
                                message = QueueMessage.model_validate_json(message_data)
                                yield message
                            except Exception as e:
                                self.logger.warning(
                                    f"Failed to parse pub/sub message: {e}"
                                )

            yield message_generator()

        finally:
            # Cleanup subscription
            try:
                await pubsub.unsubscribe()
                await pubsub.close()
            except Exception as e:
                self.logger.warning(f"Error closing pub/sub connection: {e}")

    # ========================================================================
    # Queue Management (Private Implementation)
    # ========================================================================

    async def _create_queue(
        self,
        name: str,
        **options: t.Any,
    ) -> None:
        """Create queue (private implementation).

        Redis queues are created implicitly on first use.
        This method is a no-op but included for interface compatibility.

        Args:
            name: Queue name
            **options: Backend-specific options
        """
        # Redis queues are created implicitly
        self.logger.debug(f"Queue {name} will be created on first use")

    async def _delete_queue(
        self,
        name: str,
        if_empty: bool = False,
    ) -> None:
        """Delete queue (private implementation).

        Args:
            name: Queue name
            if_empty: Only delete if empty

        Raises:
            MessagingOperationError: If deletion fails
        """
        if not self._connected:
            raise MessagingConnectionError("Not connected to Redis")

        client = await self._ensure_client()
        queue_key = self._queue_key.format(topic=name)

        try:
            if if_empty:
                size = await client.zcard(queue_key.encode())
                if size > 0:
                    raise MessagingOperationError(f"Queue {name} is not empty")

            # Get all message keys and delete
            message_keys = await client.zrange(queue_key.encode(), 0, -1)

            async with client.pipeline() as pipe:
                for message_key in message_keys:
                    await pipe.delete(message_key)
                await pipe.delete(queue_key.encode())
                await pipe.execute()

            self.logger.info(f"Deleted queue {name}")

        except Exception as e:
            self.logger.exception(f"Failed to delete queue {name}: {e}")
            raise MessagingOperationError(
                f"Failed to delete queue {name}",
                original_error=e,
            ) from e

    async def _purge_queue(
        self,
        name: str,
    ) -> int:
        """Purge queue (private implementation).

        Args:
            name: Queue name

        Returns:
            Number of messages purged

        Raises:
            MessagingOperationError: If purge fails
        """
        if not self._connected:
            raise MessagingConnectionError("Not connected to Redis")

        client = await self._ensure_client()
        queue_key = self._queue_key.format(topic=name)

        try:
            # Get all message keys
            message_keys = await client.zrange(queue_key.encode(), 0, -1)
            count = len(message_keys)

            if count > 0:
                async with client.pipeline() as pipe:
                    for message_key in message_keys:
                        await pipe.delete(message_key)
                    await pipe.delete(queue_key.encode())
                    await pipe.execute()

            self.logger.info(f"Purged {count} messages from queue {name}")
            return count

        except Exception as e:
            self.logger.exception(f"Failed to purge queue {name}: {e}")
            raise MessagingOperationError(
                f"Failed to purge queue {name}",
                original_error=e,
            ) from e

    async def _get_queue_size(
        self,
        name: str,
    ) -> int:
        """Get queue size (private implementation).

        Args:
            name: Queue name

        Returns:
            Message count

        Raises:
            MessagingOperationError: If operation fails
        """
        if not self._connected:
            raise MessagingConnectionError("Not connected to Redis")

        client = await self._ensure_client()
        queue_key = self._queue_key.format(topic=name)

        try:
            size = await client.zcard(queue_key.encode())
            return size

        except Exception as e:
            self.logger.exception(f"Failed to get queue size for {name}: {e}")
            raise MessagingOperationError(
                f"Failed to get queue size for {name}",
                original_error=e,
            ) from e

    async def _list_queues(
        self,
        pattern: str | None = None,
    ) -> list[str]:
        """List queues (private implementation).

        Args:
            pattern: Optional filter pattern

        Returns:
            List of queue names

        Raises:
            MessagingOperationError: If operation fails
        """
        if not self._connected:
            raise MessagingConnectionError("Not connected to Redis")

        client = await self._ensure_client()

        try:
            # Build key pattern
            if pattern:
                key_pattern = self._queue_key.format(topic=pattern)
            else:
                key_pattern = self._queue_key.format(topic="*")

            # Get matching keys
            keys = await client.keys(key_pattern.encode())

            # Extract queue names
            prefix = self._queue_key.format(topic="").encode()
            queue_names = [
                key[len(prefix) :].decode() for key in keys if key.startswith(prefix)
            ]

            return sorted(queue_names)

        except Exception as e:
            self.logger.exception(f"Failed to list queues: {e}")
            raise MessagingOperationError(
                "Failed to list queues",
                original_error=e,
            ) from e

    # ========================================================================
    # Background Tasks
    # ========================================================================

    async def _get_ready_delayed_messages(
        self, client: t.Any, current_time: float
    ) -> list[bytes]:
        """Get delayed messages that are ready to be processed."""
        return await client.zrangebyscore(
            self._delayed_key.encode(),
            b"-inf",
            str(current_time).encode(),
            start=0,
            num=self._settings.batch_size,
        )

    async def _move_delayed_message_to_queue(
        self, client: t.Any, message_key: bytes
    ) -> None:
        """Move a single delayed message to its target queue."""
        try:
            # Get message data
            message_data = await client.get(message_key)
            if not message_data:
                return

            message = QueueMessage.model_validate_json(message_data)

            # Calculate new score and queue key
            score = time.time() * 1_000_000 - message.priority.value
            queue_key = self._queue_key.format(topic=message.queue)

            # Move atomically
            async with client.pipeline() as pipe:
                await pipe.zrem(self._delayed_key.encode(), message_key)
                await pipe.zadd(queue_key.encode(), {message_key: score})
                await pipe.execute()

        except Exception as e:
            self.logger.warning(f"Failed to process delayed message: {e}")

    async def _process_delayed_batch(self, client: t.Any, current_time: float) -> int:
        """Process a batch of ready delayed messages. Returns count processed."""
        messages = await self._get_ready_delayed_messages(client, current_time)

        if not messages:
            return 0

        # Move messages to appropriate queues
        for message_key in messages:
            await self._move_delayed_message_to_queue(client, message_key)

        self.logger.debug(f"Processed {len(messages)} delayed messages to queues")
        return len(messages)

    async def _process_delayed_messages(self) -> None:
        """Background task to process delayed messages."""
        self.logger.debug("Started delayed message processor")

        while self._connected and not self._shutdown_event.is_set():
            try:
                client = await self._ensure_client()
                current_time = time.time() * 1_000_000

                # Process batch of ready delayed messages
                await self._process_delayed_batch(client, current_time)

                # Sleep before next check
                await asyncio.sleep(1.0)

            except asyncio.CancelledError:
                break
            except Exception as e:
                self.logger.exception(f"Error in delayed message processor: {e}")
                await asyncio.sleep(5.0)

        self.logger.debug("Stopped delayed message processor")

    # ========================================================================
    # Utility Methods
    # ========================================================================

    def get_capabilities(self) -> set[MessagingCapability]:
        """Get backend capabilities.

        Returns:
            Set of supported capabilities
        """
        capabilities = {
            MessagingCapability.BASIC_QUEUE,
            MessagingCapability.PUB_SUB,
            MessagingCapability.PRIORITY_QUEUE,
            MessagingCapability.DELAYED_MESSAGES,
            MessagingCapability.PERSISTENCE,
            MessagingCapability.DEAD_LETTER_QUEUE,
            MessagingCapability.CONNECTION_POOLING,
            MessagingCapability.MESSAGE_TTL,
            MessagingCapability.BATCH_OPERATIONS,
            MessagingCapability.PATTERN_SUBSCRIBE,
            MessagingCapability.BROADCAST,
        }

        if self._settings.enable_clustering:
            capabilities.add(MessagingCapability.CLUSTERING)

        return capabilities

    # ========================================================================
    # Public Interface Methods (Required by UnifiedMessagingBackend)
    # ========================================================================

    async def connect(self) -> None:
        """Establish connection to Redis backend."""
        await self._connect()

    async def disconnect(self) -> None:
        """Close connection to Redis backend."""
        await self._disconnect()

    async def health_check(self) -> dict[str, t.Any]:
        """Perform health check on Redis backend."""
        return await self._health_check()

    # Pub/Sub Interface (for events system)

    async def publish(
        self,
        topic: str,
        message: bytes,
        headers: dict[str, str] | None = None,
    ) -> None:
        """Publish a message to a topic (pub/sub pattern)."""
        from uuid import uuid4

        queue_msg = QueueMessage(
            message_id=uuid4(),
            queue=topic,
            payload=message,
            headers=headers or {},
        )
        await self._send(queue_msg)

    async def subscribe(
        self,
        topic: str,
        pattern: bool = False,
    ) -> Subscription:
        """Subscribe to a topic or pattern."""
        subscription = Subscription(topic=topic)
        return subscription

    async def unsubscribe(self, subscription: Subscription) -> None:
        """Unsubscribe from a topic."""
        # Cleanup handled by context manager in receive_messages
        pass

    @asynccontextmanager  # type: ignore[arg-type]
    async def receive_messages(
        self,
        subscription: Subscription,
        timeout: float | None = None,
    ) -> AsyncIterator[AsyncIterator[PubSubMessage]]:
        """Receive messages from a subscription."""
        async with self._subscribe(subscription.topic) as queue_messages:

            async def convert_messages() -> AsyncGenerator[PubSubMessage]:
                """Convert QueueMessage to PubSubMessage."""
                async for queue_msg in queue_messages:
                    pubsub_msg = PubSubMessage(
                        message_id=queue_msg.message_id,
                        topic=queue_msg.queue,
                        payload=queue_msg.payload,
                        correlation_id=queue_msg.correlation_id,
                        headers=queue_msg.headers,
                    )
                    yield pubsub_msg

            yield convert_messages()

    # Queue Interface (for tasks system)

    async def enqueue(
        self,
        queue: str,
        message: bytes,
        priority: MessagePriority = MessagePriority.NORMAL,
        delay_seconds: float = 0.0,
        headers: dict[str, str] | None = None,
    ) -> str:
        """Add a message to a queue."""
        from uuid import uuid4

        queue_msg = QueueMessage(
            message_id=uuid4(),
            queue=queue,
            payload=message,
            priority=priority,
            delay_seconds=delay_seconds,
            headers=headers or {},
        )
        return await self._send(queue_msg)

    async def dequeue(
        self,
        queue: str,
        timeout: float | None = None,
        visibility_timeout: float = 30.0,
    ) -> QueueMessage | None:
        """Remove and return a message from a queue."""
        return await self._receive(queue, timeout)

    async def acknowledge(
        self,
        queue: str,
        message_id: str,
    ) -> None:
        """Acknowledge successful processing of a message."""
        # Find the message in processing (simplified - would need tracking in real impl)
        # For now, just acknowledge by message_id
        from uuid import UUID

        dummy_msg = QueueMessage(
            message_id=UUID(message_id),
            queue=queue,
            payload=b"",
        )
        await self._acknowledge(dummy_msg)

    async def reject(
        self,
        queue: str,
        message_id: str,
        requeue: bool = True,
    ) -> None:
        """Reject a message, optionally requeuing it."""
        from uuid import UUID

        dummy_msg = QueueMessage(
            message_id=UUID(message_id),
            queue=queue,
            payload=b"",
        )
        await self._reject(dummy_msg, requeue)

    async def purge_queue(self, queue: str) -> int:
        """Remove all messages from a queue."""
        return await self._purge_queue(queue)

    async def get_queue_stats(self, queue: str) -> dict[str, t.Any]:
        """Get statistics for a queue."""
        size = await self._get_queue_size(queue)
        return {
            "message_count": size,
            "consumer_count": 0,  # Redis doesn't track consumers directly
        }


# Factory function
def create_redis_messaging(
    settings: RedisMessagingSettings | None = None,
) -> RedisMessaging:
    """Create a Redis messaging instance.

    Args:
        settings: Messaging settings

    Returns:
        RedisMessaging instance
    """
    return RedisMessaging(settings)


# Export with role-specific names for dependency injection
RedisPubSub = RedisMessaging  # For events system (pubsub adapter)
RedisQueue = RedisMessaging  # For tasks system (queue adapter)

Messaging = RedisMessaging
MessagingSettings = RedisMessagingSettings

depends.set(Messaging, "redis")



import os
import tempfile
import typing as t
from contextlib import asynccontextmanager
from functools import cached_property
from pathlib import Path
from uuid import UUID

from aioftp import AsyncPathIO, Client, Permission, Server, User
from anyio import Path as AsyncPath
from acb.adapters import AdapterCapability, AdapterMetadata, AdapterStatus
from acb.depends import depends

from ._base import FileInfo, FtpdBase, FtpdBaseSettings

MODULE_ID = UUID("0197ff55-9026-7672-b2aa-b7f5f8443c6c")
MODULE_STATUS = AdapterStatus.STABLE

MODULE_METADATA = AdapterMetadata(
    module_id=MODULE_ID,
    name="FTP Server",
    category="ftpd",
    provider="aioftp",
    version="1.0.0",
    acb_min_version="0.18.0",
    author="lesleslie <les@wedgwoodwebworks.com>",
    created_date="2025-01-12",
    last_modified="2025-01-20",
    status=MODULE_STATUS,
    capabilities=[
        AdapterCapability.ASYNC_OPERATIONS,
        AdapterCapability.CONNECTION_POOLING,
    ],
    required_packages=["aioftp", "anyio"],
    description="FTP server and client adapter with async support",
    settings_class="FtpdSettings",
    config_example={
        "host": "127.0.0.1",
        "port": 8021,
        "username": "ftpuser",
        "password": "ftppass",  # pragma: allowlist secret
        "root_dir": "/tmp/ftp",  # nosec B108
        "anonymous": False,
        "passive_ports_min": 50000,
        "passive_ports_max": 50100,
    },
)


class FtpdSettings(FtpdBaseSettings):
    port: int = 8021
    passive_ports_min: int = 50000
    passive_ports_max: int = 50100
    timeout: int = 30


class Ftpd(FtpdBase):
    _client: Client | None = None  # type: ignore[assignment]
    _path_io: AsyncPathIO | None = None

    @cached_property
    def server(self) -> Server:
        Path(self.config.ftpd.root_dir).mkdir(parents=True, exist_ok=True)
        user = User(
            login=self.config.ftpd.username,
            password=self.config.ftpd.password.get_secret_value(),
            base_path=self.config.ftpd.root_dir,
            permissions=[Permission()],
        )
        users = [user]
        if self.config.ftpd.anonymous:
            anonymous_user = User(
                login="anonymous",
                password="",  # nosec B106
                base_path=self.config.ftpd.root_dir,
                permissions=[Permission(writable=False)],
            )
            users.append(anonymous_user)
        server_kwargs: dict[str, t.Any] = {
            "users": users,
            "path_io_factory": AsyncPathIO,
            "maximum_connections": self.config.ftpd.max_connections,
        }
        if hasattr(self.config.ftpd, "host"):
            server_kwargs["host"] = self.config.ftpd.host
        if hasattr(self.config.ftpd, "port"):
            server_kwargs["port"] = self.config.ftpd.port
        if hasattr(self.config.ftpd, "host"):
            server_kwargs["passive_host"] = self.config.ftpd.host
        if hasattr(self.config.ftpd, "passive_ports_min") and hasattr(
            self.config.ftpd,
            "passive_ports_max",
        ):
            server_kwargs["passive_ports"] = range(
                self.config.ftpd.passive_ports_min,
                self.config.ftpd.passive_ports_max,
            )
        if hasattr(self.config.ftpd, "timeout"):
            server_kwargs["timeout"] = self.config.ftpd.timeout
        return Server(**server_kwargs)

    @depends.inject
    async def start(self, logger: t.Any = depends()) -> None:
        try:
            await self.server.start()
            logger.info(  # type: ignore[no-untyped-call]
                f"FTP server started on {self.config.ftpd.host}:{self.config.ftpd.port}",
            )
        except Exception as exc:
            await self.server.close()
            logger.exception(f"Error starting FTP server: {exc}")  # type: ignore[no-untyped-call]
            raise

    @depends.inject
    async def stop(self, logger: t.Any = depends()) -> None:
        try:
            await self.server.close()
            logger.info("FTP server stopped")  # type: ignore[no-untyped-call]
        except Exception as exc:
            logger.exception(f"Error stopping FTP server: {exc}")  # type: ignore[no-untyped-call]
            raise

    async def _ensure_client(self) -> Client:
        if self._client is None:
            self._client = Client()
            await self._client.connect(self.config.ftpd.host, self.config.ftpd.port)
            await self._client.login(
                self.config.ftpd.username,
                self.config.ftpd.password.get_secret_value(),
            )
        return self._client

    async def _ensure_path_io(self) -> AsyncPathIO:
        if self._path_io is None:
            self._path_io = AsyncPathIO()
        return self._path_io

    @asynccontextmanager
    async def connect(self) -> t.AsyncGenerator["Ftpd"]:
        client = await self._ensure_client()
        try:
            yield self
        finally:
            await client.quit()
            self._client = None

    async def upload(self, local_path: Path, remote_path: str) -> None:
        client = await self._ensure_client()
        await client.upload(str(local_path), remote_path)

    async def download(self, remote_path: str, local_path: Path) -> None:
        client = await self._ensure_client()
        await client.download(remote_path, str(local_path))

    async def list_dir(self, path: str) -> list[FileInfo]:
        client = await self._ensure_client()
        result = []
        async for file_info in client.list(path):
            info = FileInfo(
                name=t.cast("str", file_info[1].get("name")),
                size=t.cast("int", file_info[1].get("size", 0)),
                is_dir=file_info[1].get("type") == "dir",
                is_file=file_info[1].get("type") == "file",
                permissions=t.cast("str", file_info[1].get("permissions", "")),
                mtime=t.cast("float", file_info[1].get("modify", 0.0)),
                owner=t.cast("str", file_info[1].get("owner", "")),
                group=t.cast("str", file_info[1].get("group", "")),
            )
            result.append(info)
        return result

    async def mkdir(self, path: str) -> None:
        client = await self._ensure_client()
        await client.make_directory(path)

    async def rmdir(self, path: str, recursive: bool = False) -> None:
        client = await self._ensure_client()
        if recursive:
            files = await self.list_dir(path)
            for file in files:
                full_path = f"{path}/{file.name}"
                if file.is_dir:
                    await self.rmdir(full_path, recursive=True)
                else:
                    await self.delete(full_path)
        await client.remove_directory(path)

    async def delete(self, path: str) -> None:
        client = await self._ensure_client()
        await client.remove_file(path)

    async def rename(self, old_path: str, new_path: str) -> None:
        client = await self._ensure_client()
        await client.rename(old_path, new_path)

    async def exists(self, path: str) -> bool:
        try:
            await self.stat(path)
            return True
        except Exception:
            return False

    async def stat(self, path: str) -> FileInfo:
        client = await self._ensure_client()
        file_info = await client.stat(path)
        return FileInfo(
            name=os.path.basename(path),
            size=int(file_info.get("size", 0)),
            is_dir=file_info["type"] == "dir",
            is_file=file_info["type"] == "file",
            permissions=str(file_info.get("permissions", "")),
            mtime=float(file_info.get("modify", 0.0)),
            owner=str(file_info.get("owner", "")),
            group=str(file_info.get("group", "")),
        )

    async def read_text(self, path: str) -> str:
        temp_dir = tempfile.mkdtemp()
        temp_path = Path(temp_dir) / os.path.basename(path)
        try:
            await self.download(path, temp_path)
            async_path = AsyncPath(temp_path)
            return await async_path.read_text()
        finally:
            if temp_path.exists():
                temp_path.unlink()
            if Path(temp_dir).exists():
                os.rmdir(temp_dir)

    async def read_bytes(self, path: str) -> bytes:
        temp_dir = tempfile.mkdtemp()
        temp_path = Path(temp_dir) / os.path.basename(path)
        try:
            await self.download(path, temp_path)
            async_path = AsyncPath(temp_path)
            return await async_path.read_bytes()
        finally:
            if temp_path.exists():
                temp_path.unlink()
            if Path(temp_dir).exists():
                os.rmdir(temp_dir)

    async def write_text(self, path: str, content: str) -> None:
        temp_dir = tempfile.mkdtemp()
        temp_path = Path(temp_dir) / os.path.basename(path)
        try:
            async_path = AsyncPath(temp_path)
            await async_path.write_text(content)
            await self.upload(temp_path, path)
        finally:
            if temp_path.exists():
                temp_path.unlink()
            if Path(temp_dir).exists():
                os.rmdir(temp_dir)

    async def write_bytes(self, path: str, content: bytes) -> None:
        temp_dir = tempfile.mkdtemp()
        temp_path = Path(temp_dir) / os.path.basename(path)
        try:
            async_path = AsyncPath(temp_path)
            await async_path.write_bytes(content)
            await self.upload(temp_path, path)
        finally:
            if temp_path.exists():
                temp_path.unlink()
            if Path(temp_dir).exists():
                os.rmdir(temp_dir)


depends.set(Ftpd, "ftp")
